class Type(MemberInfo,ICustomAttributeProvider,_MemberInfo,_Type,IReflect):
 """ Represents type declarations: class types,interface types,array types,value types,enumeration types,type parameters,generic type definitions,and open or closed constructed generic types. """
 def Equals(self,o):
  """
  Equals(self: Type,o: Type) -> bool

  

   Determines if the underlying system type of the current System.Type is the same as the 

    underlying system type of the specified System.Type.

  

  

   o: The object whose underlying system type is to be compared with the underlying system type of the 

    current System.Type.

  

   Returns: true if the underlying system type of o is the same as the underlying system type of the current 

    System.Type; otherwise,false.

  

  Equals(self: Type,o: object) -> bool

  

   Determines if the underlying system type of the current System.Type is the same as the 

    underlying system type of the specified System.Object.

  

  

   o: The object whose underlying system type is to be compared with the underlying system type of the 

    current System.Type.

  

   Returns: true if the underlying system type of o is the same as the underlying system type of the current 

    System.Type; otherwise,false. This method also returns false if the object specified by the o 

    parameter is not a Type.
  """
  pass
 def FilterAttribute(self,*args):
  """
  Represents a delegate that is used to filter a list of members represented in an array of System.Reflection.MemberInfo objects.

  

  MemberFilter(object: object,method: IntPtr)
  """
  pass
 def FilterName(self,*args):
  """
  Represents a delegate that is used to filter a list of members represented in an array of System.Reflection.MemberInfo objects.

  

  MemberFilter(object: object,method: IntPtr)
  """
  pass
 def FilterNameIgnoreCase(self,*args):
  """
  Represents a delegate that is used to filter a list of members represented in an array of System.Reflection.MemberInfo objects.

  

  MemberFilter(object: object,method: IntPtr)
  """
  pass
 def FindInterfaces(self,filter,filterCriteria):
  """
  FindInterfaces(self: Type,filter: TypeFilter,filterCriteria: object) -> Array[Type]

  

   Returns an array of System.Type objects representing a filtered list of interfaces implemented 

    or inherited by the current System.Type.

  

  

   filter: The delegate that compares the interfaces against filterCriteria.

   filterCriteria: The search criteria that determines whether an interface should be included in the returned 

    array.

  

   Returns: An array of System.Type objects representing a filtered list of the interfaces implemented or 

    inherited by the current System.Type,or an empty array of type System.Type if no interfaces 

    matching the filter are implemented or inherited by the current System.Type.
  """
  pass
 def FindMembers(self,memberType,bindingAttr,filter,filterCriteria):
  """
  FindMembers(self: Type,memberType: MemberTypes,bindingAttr: BindingFlags,filter: MemberFilter,filterCriteria: object) -> Array[MemberInfo]

  

   Returns a filtered array of System.Reflection.MemberInfo objects of the specified member type.

  

   memberType: An object that indicates the type of member to search for.

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   filter: The delegate that does the comparisons,returning true if the member currently being inspected 

    matches the filterCriteria and false otherwise. You can use the FilterAttribute,FilterName,and 

    FilterNameIgnoreCase delegates supplied by this class. The first uses the fields of 

    FieldAttributes,MethodAttributes,and MethodImplAttributes as search criteria,and the other 

    two delegates use String objects as the search criteria.

  

   filterCriteria: The search criteria that determines whether a member is returned in the array of MemberInfo 

    objects.The fields of FieldAttributes,MethodAttributes,and MethodImplAttributes can be used in 

    conjunction with the FilterAttribute delegate supplied by this class.

  

   Returns: A filtered array of System.Reflection.MemberInfo objects of the specified member type.-or- An 

    empty array of type System.Reflection.MemberInfo,if the current System.Type does not have 

    members of type memberType that match the filter criteria.
  """
  pass
 def GetArrayRank(self):
  """
  GetArrayRank(self: Type) -> int

  

   Gets the number of dimensions in an System.Array.

   Returns: An System.Int32 containing the number of dimensions in the current Type.
  """
  pass
 def GetAttributeFlagsImpl(self,*args):
  """
  GetAttributeFlagsImpl(self: Type) -> TypeAttributes

  

   When overridden in a derived class,implements the System.Type.Attributes property and gets a 

    bitmask indicating the attributes associated with the System.Type.

  

   Returns: A System.Reflection.TypeAttributes object representing the attribute set of the System.Type.
  """
  pass
 def GetConstructor(self,*__args):
  """
  GetConstructor(self: Type,types: Array[Type]) -> ConstructorInfo

  

   Searches for a public instance constructor whose parameters match the types in the specified 

    array.

  

  

   types: An array of System.Type objects representing the number,order,and type of the parameters for 

    the desired constructor.-or- An empty array of System.Type objects,to get a constructor that 

    takes no parameters. Such an empty array is provided by the static field System.Type.EmptyTypes.

  

   Returns: An object representing the public instance constructor whose parameters match the types in the 

    parameter type array,if found; otherwise,null.

  

  GetConstructor(self: Type,bindingAttr: BindingFlags,binder: Binder,types: Array[Type],modifiers: Array[ParameterModifier]) -> ConstructorInfo

  

   Searches for a constructor whose parameters match the specified argument types and modifiers,

    using the specified binding constraints.

  

  

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   binder: An object that defines a set of properties and enables binding,which can involve selection of 

    an overloaded method,coercion of argument types,and invocation of a member through 

    reflection.-or- A null reference (Nothing in Visual Basic),to use the 

    System.Type.DefaultBinder.

  

   types: An array of System.Type objects representing the number,order,and type of the parameters for 

    the constructor to get.-or- An empty array of the type System.Type (that is,Type[] types=new 

    Type[0]) to get a constructor that takes no parameters.-or- System.Type.EmptyTypes.

  

   modifiers: An array of System.Reflection.ParameterModifier objects representing the attributes associated 

    with the corresponding element in the parameter type array. The default binder does not process 

    this parameter.

  

   Returns: A System.Reflection.ConstructorInfo object representing the constructor that matches the 

    specified requirements,if found; otherwise,null.

  

  GetConstructor(self: Type,bindingAttr: BindingFlags,binder: Binder,callConvention: CallingConventions,types: Array[Type],modifiers: Array[ParameterModifier]) -> ConstructorInfo

  

   Searches for a constructor whose parameters match the specified argument types and modifiers,

    using the specified binding constraints and the specified calling convention.

  

  

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   binder: An object that defines a set of properties and enables binding,which can involve selection of 

    an overloaded method,coercion of argument types,and invocation of a member through 

    reflection.-or- A null reference (Nothing in Visual Basic),to use the 

    System.Type.DefaultBinder.

  

   callConvention: The object that specifies the set of rules to use regarding the order and layout of arguments,

    how the return value is passed,what registers are used for arguments,and the stack is cleaned 

    up.

  

   types: An array of System.Type objects representing the number,order,and type of the parameters for 

    the constructor to get.-or- An empty array of the type System.Type (that is,Type[] types=new 

    Type[0]) to get a constructor that takes no parameters.

  

   modifiers: An array of System.Reflection.ParameterModifier objects representing the attributes associated 

    with the corresponding element in the types array. The default binder does not process this 

    parameter.

  

   Returns: An object representing the constructor that matches the specified requirements,if found; 

    otherwise,null.
  """
  pass
 def GetConstructorImpl(self,*args):
  """
  GetConstructorImpl(self: Type,bindingAttr: BindingFlags,binder: Binder,callConvention: CallingConventions,types: Array[Type],modifiers: Array[ParameterModifier]) -> ConstructorInfo

  

   When overridden in a derived class,searches for a constructor whose parameters match the 

    specified argument types and modifiers,using the specified binding constraints and the 

    specified calling convention.

  

  

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   binder: An object that defines a set of properties and enables binding,which can involve selection of 

    an overloaded method,coercion of argument types,and invocation of a member through 

    reflection.-or- A null reference (Nothing in Visual Basic),to use the 

    System.Type.DefaultBinder.

  

   callConvention: The object that specifies the set of rules to use regarding the order and layout of arguments,

    how the return value is passed,what registers are used for arguments,and the stack is cleaned 

    up.

  

   types: An array of System.Type objects representing the number,order,and type of the parameters for 

    the constructor to get.-or- An empty array of the type System.Type (that is,Type[] types=new 

    Type[0]) to get a constructor that takes no parameters.

  

   modifiers: An array of System.Reflection.ParameterModifier objects representing the attributes associated 

    with the corresponding element in the types array. The default binder does not process this 

    parameter.

  

   Returns: A System.Reflection.ConstructorInfo object representing the constructor that matches the 

    specified requirements,if found; otherwise,null.
  """
  pass
 def GetConstructors(self,bindingAttr=None):
  """
  GetConstructors(self: Type) -> Array[ConstructorInfo]

  

   Returns all the public constructors defined for the current System.Type.

   Returns: An array of System.Reflection.ConstructorInfo objects representing all the public instance 

    constructors defined for the current System.Type,but not including the type initializer (static 

    constructor). If no public instance constructors are defined for the current System.Type,or if 

    the current System.Type represents a type parameter in the definition of a generic type or 

    generic method,an empty array of type System.Reflection.ConstructorInfo is returned.

  

  GetConstructors(self: Type,bindingAttr: BindingFlags) -> Array[ConstructorInfo]

  

   When overridden in a derived class,searches for the constructors defined for the current 

    System.Type,using the specified BindingFlags.

  

  

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   Returns: An array of System.Reflection.ConstructorInfo objects representing all constructors defined for 

    the current System.Type that match the specified binding constraints,including the type 

    initializer if it is defined. Returns an empty array of type System.Reflection.ConstructorInfo 

    if no constructors are defined for the current System.Type,if none of the defined constructors 

    match the binding constraints,or if the current System.Type represents a type parameter in the 

    definition of a generic type or generic method.
  """
  pass
 def GetDefaultMembers(self):
  """
  GetDefaultMembers(self: Type) -> Array[MemberInfo]

  

   Searches for the members defined for the current System.Type whose 

    System.Reflection.DefaultMemberAttribute is set.

  

   Returns: An array of System.Reflection.MemberInfo objects representing all default members of the current 

    System.Type.-or- An empty array of type System.Reflection.MemberInfo,if the current System.Type 

    does not have default members.
  """
  pass
 def GetElementType(self):
  """
  GetElementType(self: Type) -> Type

  

   When overridden in a derived class,returns the System.Type of the object encompassed or 

    referred to by the current array,pointer or reference type.

  

   Returns: The System.Type of the object encompassed or referred to by the current array,pointer,or 

    reference type,or null if the current System.Type is not an array or a pointer,or is not 

    passed by reference,or represents a generic type or a type parameter in the definition of a 

    generic type or generic method.
  """
  pass
 def GetEnumName(self,value):
  """
  GetEnumName(self: Type,value: object) -> str

  

   Returns the name of the constant that has the specified value,for the current enumeration type.

  

   value: The value whose name is to be retrieved.

   Returns: The name of the member of the current enumeration type that has the specified value,or null if 

    no such constant is found.
  """
  pass
 def GetEnumNames(self):
  """
  GetEnumNames(self: Type) -> Array[str]

  

   Returns the names of the members of the current enumeration type.

   Returns: An array that contains the names of the members of the enumeration.
  """
  pass
 def GetEnumUnderlyingType(self):
  """
  GetEnumUnderlyingType(self: Type) -> Type

  

   Returns the underlying type of the current enumeration type.

   Returns: The underlying type of the current enumeration.
  """
  pass
 def GetEnumValues(self):
  """
  GetEnumValues(self: Type) -> Array

  

   Returns an array of the values of the constants in the current enumeration type.

   Returns: An array that contains the values. The elements of the array are sorted by the binary values 

    (that is,the unsigned values) of the enumeration constants.
  """
  pass
 def GetEvent(self,name,bindingAttr=None):
  """
  GetEvent(self: Type,name: str,bindingAttr: BindingFlags) -> EventInfo

  

   When overridden in a derived class,returns the System.Reflection.EventInfo object representing 

    the specified event,using the specified binding constraints.

  

  

   name: The string containing the name of an event which is declared or inherited by the current 

    System.Type.

  

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   Returns: The object representing the specified event that is declared or inherited by the current 

    System.Type,if found; otherwise,null.

  

  GetEvent(self: Type,name: str) -> EventInfo

  

   Returns the System.Reflection.EventInfo object representing the specified public event.

  

   name: The string containing the name of an event that is declared or inherited by the current 

    System.Type.

  

   Returns: The object representing the specified public event that is declared or inherited by the current 

    System.Type,if found; otherwise,null.
  """
  pass
 def GetEvents(self,bindingAttr=None):
  """
  GetEvents(self: Type,bindingAttr: BindingFlags) -> Array[EventInfo]

  

   When overridden in a derived class,searches for events that are declared or inherited by the 

    current System.Type,using the specified binding constraints.

  

  

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   Returns: An array of System.Reflection.EventInfo objects representing all events that are declared or 

    inherited by the current System.Type that match the specified binding constraints.-or- An empty 

    array of type System.Reflection.EventInfo,if the current System.Type does not have events,or 

    if none of the events match the binding constraints.

  

  GetEvents(self: Type) -> Array[EventInfo]

  

   Returns all the public events that are declared or inherited by the current System.Type.

   Returns: An array of System.Reflection.EventInfo objects representing all the public events which are 

    declared or inherited by the current System.Type.-or- An empty array of type 

    System.Reflection.EventInfo,if the current System.Type does not have public events.
  """
  pass
 def GetField(self,name,bindingAttr=None):
  """
  GetField(self: Type,name: str) -> FieldInfo

  

   Searches for the public field with the specified name.

  

   name: The string containing the name of the data field to get.

   Returns: An object representing the public field with the specified name,if found; otherwise,null.

  GetField(self: Type,name: str,bindingAttr: BindingFlags) -> FieldInfo

  

   Searches for the specified field,using the specified binding constraints.

  

   name: The string containing the name of the data field to get.

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   Returns: An object representing the field that matches the specified requirements,if found; otherwise,

    null.
  """
  pass
 def GetFields(self,bindingAttr=None):
  """
  GetFields(self: Type,bindingAttr: BindingFlags) -> Array[FieldInfo]

  

   When overridden in a derived class,searches for the fields defined for the current System.Type,

    using the specified binding constraints.

  

  

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   Returns: An array of System.Reflection.FieldInfo objects representing all fields defined for the current 

    System.Type that match the specified binding constraints.-or- An empty array of type 

    System.Reflection.FieldInfo,if no fields are defined for the current System.Type,or if none of 

    the defined fields match the binding constraints.

  

  GetFields(self: Type) -> Array[FieldInfo]

  

   Returns all the public fields of the current System.Type.

   Returns: An array of System.Reflection.FieldInfo objects representing all the public fields defined for 

    the current System.Type.-or- An empty array of type System.Reflection.FieldInfo,if no public 

    fields are defined for the current System.Type.
  """
  pass
 def GetGenericArguments(self):
  """
  GetGenericArguments(self: Type) -> Array[Type]

  

   Returns an array of System.Type objects that represent the type arguments of a generic type or 

    the type parameters of a generic type definition.

  

   Returns: An array of System.Type objects that represent the type arguments of a generic type. Returns an 

    empty array if the current type is not a generic type.
  """
  pass
 def GetGenericParameterConstraints(self):
  """
  GetGenericParameterConstraints(self: Type) -> Array[Type]

  

   Returns an array of System.Type objects that represent the constraints on the current generic 

    type parameter.

  

   Returns: An array of System.Type objects that represent the constraints on the current generic type 

    parameter.
  """
  pass
 def GetGenericTypeDefinition(self):
  """
  GetGenericTypeDefinition(self: Type) -> Type

  

   Returns a System.Type object that represents a generic type definition from which the current 

    generic type can be constructed.

  

   Returns: A System.Type object representing a generic type from which the current type can be constructed.
  """
  pass
 def GetHashCode(self):
  """
  GetHashCode(self: Type) -> int

  

   Returns the hash code for this instance.

   Returns: The hash code for this instance.
  """
  pass
 def GetInterface(self,name,ignoreCase=None):
  """
  GetInterface(self: Type,name: str,ignoreCase: bool) -> Type

  

   When overridden in a derived class,searches for the specified interface,specifying whether to 

    do a case-insensitive search for the interface name.

  

  

   name: The string containing the name of the interface to get. For generic interfaces,this is the 

    mangled name.

  

   ignoreCase: true to ignore the case of that part of name that specifies the simple interface name (the part 

    that specifies the namespace must be correctly cased).-or- false to perform a case-sensitive 

    search for all parts of name.

  

   Returns: An object representing the interface with the specified name,implemented or inherited by the 

    current System.Type,if found; otherwise,null.

  

  GetInterface(self: Type,name: str) -> Type

  

   Searches for the interface with the specified name.

  

   name: The string containing the name of the interface to get. For generic interfaces,this is the 

    mangled name.

  

   Returns: An object representing the interface with the specified name,implemented or inherited by the 

    current System.Type,if found; otherwise,null.
  """
  pass
 def GetInterfaceMap(self,interfaceType):
  """
  GetInterfaceMap(self: Type,interfaceType: Type) -> InterfaceMapping

  

   Returns an interface mapping for the specified interface type.

  

   interfaceType: The interface type to retrieve a mapping for.

   Returns: An object that represents the interface mapping for interfaceType.
  """
  pass
 def GetInterfaces(self):
  """
  GetInterfaces(self: Type) -> Array[Type]

  

   When overridden in a derived class,gets all the interfaces implemented or inherited by the 

    current System.Type.

  

   Returns: An array of System.Type objects representing all the interfaces implemented or inherited by the 

    current System.Type.-or- An empty array of type System.Type,if no interfaces are implemented or 

    inherited by the current System.Type.
  """
  pass
 def GetMember(self,name,*__args):
  """
  GetMember(self: Type,name: str,type: MemberTypes,bindingAttr: BindingFlags) -> Array[MemberInfo]

  

   Searches for the specified members of the specified member type,using the specified binding 

    constraints.

  

  

   name: The string containing the name of the members to get.

   type: The value to search for.

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return an empty array.

  

   Returns: An array of System.Reflection.MemberInfo objects representing the public members with the 

    specified name,if found; otherwise,an empty array.

  

  GetMember(self: Type,name: str,bindingAttr: BindingFlags) -> Array[MemberInfo]

  

   Searches for the specified members,using the specified binding constraints.

  

   name: The string containing the name of the members to get.

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return an empty array.

  

   Returns: An array of System.Reflection.MemberInfo objects representing the public members with the 

    specified name,if found; otherwise,an empty array.

  

  GetMember(self: Type,name: str) -> Array[MemberInfo]

  

   Searches for the public members with the specified name.

  

   name: The string containing the name of the public members to get.

   Returns: An array of System.Reflection.MemberInfo objects representing the public members with the 

    specified name,if found; otherwise,an empty array.
  """
  pass
 def GetMembers(self,bindingAttr=None):
  """
  GetMembers(self: Type,bindingAttr: BindingFlags) -> Array[MemberInfo]

  

   When overridden in a derived class,searches for the members defined for the current 

    System.Type,using the specified binding constraints.

  

  

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   Returns: An array of System.Reflection.MemberInfo objects representing all members defined for the 

    current System.Type that match the specified binding constraints.-or- An empty array of type 

    System.Reflection.MemberInfo,if no members are defined for the current System.Type,or if none 

    of the defined members match the binding constraints.

  

  GetMembers(self: Type) -> Array[MemberInfo]

  

   Returns all the public members of the current System.Type.

   Returns: An array of System.Reflection.MemberInfo objects representing all the public members of the 

    current System.Type.-or- An empty array of type System.Reflection.MemberInfo,if the current 

    System.Type does not have public members.
  """
  pass
 def GetMethod(self,name,*__args):
  """
  GetMethod(self: Type,name: str,types: Array[Type]) -> MethodInfo

  

   Searches for the specified public method whose parameters match the specified argument types.

  

   name: The string containing the name of the public method to get.

   types: An array of System.Type objects representing the number,order,and type of the parameters for 

    the method to get.-or- An empty array of System.Type objects (as provided by the 

    System.Type.EmptyTypes field) to get a method that takes no parameters.

  

   Returns: An object representing the public method whose parameters match the specified argument types,if 

    found; otherwise,null.

  

  GetMethod(self: Type,name: str,bindingAttr: BindingFlags) -> MethodInfo

  

   Searches for the specified method,using the specified binding constraints.

  

   name: The string containing the name of the method to get.

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   Returns: An object representing the method that matches the specified requirements,if found; otherwise,

    null.

  

  GetMethod(self: Type,name: str) -> MethodInfo

  

   Searches for the public method with the specified name.

  

   name: The string containing the name of the public method to get.

   Returns: An object that represents the public method with the specified name,if found; otherwise,null.

  GetMethod(self: Type,name: str,bindingAttr: BindingFlags,binder: Binder,callConvention: CallingConventions,types: Array[Type],modifiers: Array[ParameterModifier]) -> MethodInfo

  

   Searches for the specified method whose parameters match the specified argument types and 

    modifiers,using the specified binding constraints and the specified calling convention.

  

  

   name: The string containing the name of the method to get.

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   binder: An object that defines a set of properties and enables binding,which can involve selection of 

    an overloaded method,coercion of argument types,and invocation of a member through 

    reflection.-or- A null reference (Nothing in Visual Basic),to use the 

    System.Type.DefaultBinder.

  

   callConvention: The object that specifies the set of rules to use regarding the order and layout of arguments,

    how the return value is passed,what registers are used for arguments,and how the stack is 

    cleaned up.

  

   types: An array of System.Type objects representing the number,order,and type of the parameters for 

    the method to get.-or- An empty array of System.Type objects (as provided by the 

    System.Type.EmptyTypes field) to get a method that takes no parameters.

  

   modifiers: An array of System.Reflection.ParameterModifier objects representing the attributes associated 

    with the corresponding element in the types array. To be only used when calling through COM 

    interop,and only parameters that are passed by reference are handled. The default binder does 

    not process this parameter.

  

   Returns: An object representing the method that matches the specified requirements,if found; otherwise,

    null.

  

  GetMethod(self: Type,name: str,bindingAttr: BindingFlags,binder: Binder,types: Array[Type],modifiers: Array[ParameterModifier]) -> MethodInfo

  

   Searches for the specified method whose parameters match the specified argument types and 

    modifiers,using the specified binding constraints.

  

  

   name: The string containing the name of the method to get.

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   binder: An object that defines a set of properties and enables binding,which can involve selection of 

    an overloaded method,coercion of argument types,and invocation of a member through 

    reflection.-or- A null reference (Nothing in Visual Basic),to use the 

    System.Type.DefaultBinder.

  

   types: An array of System.Type objects representing the number,order,and type of the parameters for 

    the method to get.-or- An empty array of System.Type objects (as provided by the 

    System.Type.EmptyTypes field) to get a method that takes no parameters.

  

   modifiers: An array of System.Reflection.ParameterModifier objects representing the attributes associated 

    with the corresponding element in the types array. To be only used when calling through COM 

    interop,and only parameters that are passed by reference are handled. The default binder does 

    not process this parameter.

  

   Returns: An object representing the method that matches the specified requirements,if found; otherwise,

    null.

  

  GetMethod(self: Type,name: str,types: Array[Type],modifiers: Array[ParameterModifier]) -> MethodInfo

  

   Searches for the specified public method whose parameters match the specified argument types and 

    modifiers.

  

  

   name: The string containing the name of the public method to get.

   types: An array of System.Type objects representing the number,order,and type of the parameters for 

    the method to get.-or- An empty array of System.Type objects (as provided by the 

    System.Type.EmptyTypes field) to get a method that takes no parameters.

  

   modifiers: An array of System.Reflection.ParameterModifier objects representing the attributes associated 

    with the corresponding element in the types array. To be only used when calling through COM 

    interop,and only parameters that are passed by reference are handled. The default binder does 

    not process this parameter.

  

   Returns: An object representing the public method that matches the specified requirements,if found; 

    otherwise,null.
  """
  pass
 def GetMethodImpl(self,*args):
  """
  GetMethodImpl(self: Type,name: str,bindingAttr: BindingFlags,binder: Binder,callConvention: CallingConventions,types: Array[Type],modifiers: Array[ParameterModifier]) -> MethodInfo

  

   When overridden in a derived class,searches for the specified method whose parameters match the 

    specified argument types and modifiers,using the specified binding constraints and the 

    specified calling convention.

  

  

   name: The string containing the name of the method to get.

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   binder: An object that defines a set of properties and enables binding,which can involve selection of 

    an overloaded method,coercion of argument types,and invocation of a member through 

    reflection.-or- A null reference (Nothing in Visual Basic),to use the 

    System.Type.DefaultBinder.

  

   callConvention: The object that specifies the set of rules to use regarding the order and layout of arguments,

    how the return value is passed,what registers are used for arguments,and what process cleans 

    up the stack.

  

   types: An array of System.Type objects representing the number,order,and type of the parameters for 

    the method to get.-or- An empty array of the type System.Type (that is,Type[] types=new 

    Type[0]) to get a method that takes no parameters.-or- null. If types is null,arguments are not 

    matched.

  

   modifiers: An array of System.Reflection.ParameterModifier objects representing the attributes associated 

    with the corresponding element in the types array. The default binder does not process this 

    parameter.

  

   Returns: An object representing the method that matches the specified requirements,if found; otherwise,

    null.
  """
  pass
 def GetMethods(self,bindingAttr=None):
  """
  GetMethods(self: Type) -> Array[MethodInfo]

  

   Returns all the public methods of the current System.Type.

   Returns: An array of System.Reflection.MethodInfo objects representing all the public methods defined for 

    the current System.Type.-or- An empty array of type System.Reflection.MethodInfo,if no public 

    methods are defined for the current System.Type.

  

  GetMethods(self: Type,bindingAttr: BindingFlags) -> Array[MethodInfo]

  

   When overridden in a derived class,searches for the methods defined for the current 

    System.Type,using the specified binding constraints.

  

  

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   Returns: An array of System.Reflection.MethodInfo objects representing all methods defined for the 

    current System.Type that match the specified binding constraints.-or- An empty array of type 

    System.Reflection.MethodInfo,if no methods are defined for the current System.Type,or if none 

    of the defined methods match the binding constraints.
  """
  pass
 def GetNestedType(self,name,bindingAttr=None):
  """
  GetNestedType(self: Type,name: str,bindingAttr: BindingFlags) -> Type

  

   When overridden in a derived class,searches for the specified nested type,using the specified 

    binding constraints.

  

  

   name: The string containing the name of the nested type to get.

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   Returns: An object representing the nested type that matches the specified requirements,if found; 

    otherwise,null.

  

  GetNestedType(self: Type,name: str) -> Type

  

   Searches for the public nested type with the specified name.

  

   name: The string containing the name of the nested type to get.

   Returns: An object representing the public nested type with the specified name,if found; otherwise,null.
  """
  pass
 def GetNestedTypes(self,bindingAttr=None):
  """
  GetNestedTypes(self: Type,bindingAttr: BindingFlags) -> Array[Type]

  

   When overridden in a derived class,searches for the types nested in the current System.Type,

    using the specified binding constraints.

  

  

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   Returns: An array of System.Type objects representing all the types nested in the current System.Type 

    that match the specified binding constraints (the search is not recursive),or an empty array of 

    type System.Type,if no nested types are found that match the binding constraints.

  

  GetNestedTypes(self: Type) -> Array[Type]

  

   Returns the public types nested in the current System.Type.

   Returns: An array of System.Type objects representing the public types nested in the current System.Type 

    (the search is not recursive),or an empty array of type System.Type if no public types are 

    nested in the current System.Type.
  """
  pass
 def GetProperties(self,bindingAttr=None):
  """
  GetProperties(self: Type) -> Array[PropertyInfo]

  

   Returns all the public properties of the current System.Type.

   Returns: An array of System.Reflection.PropertyInfo objects representing all public properties of the 

    current System.Type.-or- An empty array of type System.Reflection.PropertyInfo,if the current 

    System.Type does not have public properties.

  

  GetProperties(self: Type,bindingAttr: BindingFlags) -> Array[PropertyInfo]

  

   When overridden in a derived class,searches for the properties of the current System.Type,

    using the specified binding constraints.

  

  

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   Returns: An array of System.Reflection.PropertyInfo objects representing all properties of the current 

    System.Type that match the specified binding constraints.-or- An empty array of type 

    System.Reflection.PropertyInfo,if the current System.Type does not have properties,or if none 

    of the properties match the binding constraints.
  """
  pass
 def GetProperty(self,name,*__args):
  """
  GetProperty(self: Type,name: str,types: Array[Type]) -> PropertyInfo

  

   Searches for the specified public property whose parameters match the specified argument types.

  

   name: The string containing the name of the public property to get.

   types: An array of System.Type objects representing the number,order,and type of the parameters for 

    the indexed property to get.-or- An empty array of the type System.Type (that is,Type[] types=

    new Type[0]) to get a property that is not indexed.

  

   Returns: An object representing the public property whose parameters match the specified argument types,

    if found; otherwise,null.

  

  GetProperty(self: Type,name: str,returnType: Type) -> PropertyInfo

  

   Searches for the public property with the specified name and return type.

  

   name: The string containing the name of the public property to get.

   returnType: The return type of the property.

   Returns: An object representing the public property with the specified name,if found; otherwise,null.

  GetProperty(self: Type,name: str) -> PropertyInfo

  

   Searches for the public property with the specified name.

  

   name: The string containing the name of the public property to get.

   Returns: An object representing the public property with the specified name,if found; otherwise,null.

  GetProperty(self: Type,name: str,returnType: Type,types: Array[Type]) -> PropertyInfo

  

   Searches for the specified public property whose parameters match the specified argument types.

  

   name: The string containing the name of the public property to get.

   returnType: The return type of the property.

   types: An array of System.Type objects representing the number,order,and type of the parameters for 

    the indexed property to get.-or- An empty array of the type System.Type (that is,Type[] types=

    new Type[0]) to get a property that is not indexed.

  

   Returns: An object representing the public property whose parameters match the specified argument types,

    if found; otherwise,null.

  

  GetProperty(self: Type,name: str,bindingAttr: BindingFlags,binder: Binder,returnType: Type,types: Array[Type],modifiers: Array[ParameterModifier]) -> PropertyInfo

  

   Searches for the specified property whose parameters match the specified argument types and 

    modifiers,using the specified binding constraints.

  

  

   name: The string containing the name of the property to get.

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   binder: An object that defines a set of properties and enables binding,which can involve selection of 

    an overloaded method,coercion of argument types,and invocation of a member through 

    reflection.-or- A null reference (Nothing in Visual Basic),to use the 

    System.Type.DefaultBinder.

  

   returnType: The return type of the property.

   types: An array of System.Type objects representing the number,order,and type of the parameters for 

    the indexed property to get.-or- An empty array of the type System.Type (that is,Type[] types=

    new Type[0]) to get a property that is not indexed.

  

   modifiers: An array of System.Reflection.ParameterModifier objects representing the attributes associated 

    with the corresponding element in the types array. The default binder does not process this 

    parameter.

  

   Returns: An object representing the property that matches the specified requirements,if found; 

    otherwise,null.

  

  GetProperty(self: Type,name: str,returnType: Type,types: Array[Type],modifiers: Array[ParameterModifier]) -> PropertyInfo

  

   Searches for the specified public property whose parameters match the specified argument types 

    and modifiers.

  

  

   name: The string containing the name of the public property to get.

   returnType: The return type of the property.

   types: An array of System.Type objects representing the number,order,and type of the parameters for 

    the indexed property to get.-or- An empty array of the type System.Type (that is,Type[] types=

    new Type[0]) to get a property that is not indexed.

  

   modifiers: An array of System.Reflection.ParameterModifier objects representing the attributes associated 

    with the corresponding element in the types array. The default binder does not process this 

    parameter.

  

   Returns: An object representing the public property that matches the specified requirements,if found; 

    otherwise,null.

  

  GetProperty(self: Type,name: str,bindingAttr: BindingFlags) -> PropertyInfo

  

   Searches for the specified property,using the specified binding constraints.

  

   name: The string containing the name of the property to get.

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   Returns: An object representing the property that matches the specified requirements,if found; 

    otherwise,null.
  """
  pass
 def GetPropertyImpl(self,*args):
  """
  GetPropertyImpl(self: Type,name: str,bindingAttr: BindingFlags,binder: Binder,returnType: Type,types: Array[Type],modifiers: Array[ParameterModifier]) -> PropertyInfo

  

   When overridden in a derived class,searches for the specified property whose parameters match 

    the specified argument types and modifiers,using the specified binding constraints.

  

  

   name: The string containing the name of the property to get.

   bindingAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted.-or- Zero,to return null.

  

   binder: An object that defines a set of properties and enables binding,which can involve selection of 

    an overloaded member,coercion of argument types,and invocation of a member through 

    reflection.-or- A null reference (Nothing in Visual Basic),to use the 

    System.Type.DefaultBinder.

  

   returnType: The return type of the property.

   types: An array of System.Type objects representing the number,order,and type of the parameters for 

    the indexed property to get.-or- An empty array of the type System.Type (that is,Type[] types=

    new Type[0]) to get a property that is not indexed.

  

   modifiers: An array of System.Reflection.ParameterModifier objects representing the attributes associated 

    with the corresponding element in the types array. The default binder does not process this 

    parameter.

  

   Returns: An object representing the property that matches the specified requirements,if found; 

    otherwise,null.
  """
  pass
 def GetType(self,typeName=None,*__args):
  """
  GetType(typeName: str,assemblyResolver: Func[AssemblyName,Assembly],typeResolver: Func[Assembly,str,bool,Type],throwOnError: bool) -> Type

  GetType(typeName: str,assemblyResolver: Func[AssemblyName,Assembly],typeResolver: Func[Assembly,str,bool,Type],throwOnError: bool,ignoreCase: bool) -> Type

  GetType(self: Type) -> Type

  

   Gets the current System.Type.

   Returns: The current System.Type.

  GetType(typeName: str,assemblyResolver: Func[AssemblyName,Assembly],typeResolver: Func[Assembly,str,bool,Type]) -> Type

  GetType(typeName: str,throwOnError: bool,ignoreCase: bool) -> Type

  

   Gets the System.Type with the specified name,specifying whether to perform a case-sensitive 

    search and whether to throw an exception if the type is not found.

  

  

   typeName: The assembly-qualified name of the type to get. See System.Type.AssemblyQualifiedName. If the 

    type is in the currently executing assembly or in Mscorlib.dll,it is sufficient to supply the 

    type name qualified by its namespace.

  

   throwOnError: true to throw an exception if the type cannot be found; false to return null.Specifying false 

    also suppresses some other exception conditions,but not all of them. See the Exceptions 

    section.

  

   ignoreCase: true to perform a case-insensitive search for typeName,false to perform a case-sensitive search 

    for typeName.

  

   Returns: The type with the specified name. If the type is not found,the throwOnError parameter specifies 

    whether null is returned or an exception is thrown. In some cases,an exception is thrown 

    regardless of the value of throwOnError. See the Exceptions section.

  

  GetType(typeName: str,throwOnError: bool) -> Type

  

   Gets the System.Type with the specified name,performing a case-sensitive search and specifying 

    whether to throw an exception if the type is not found.

  

  

   typeName: The assembly-qualified name of the type to get. See System.Type.AssemblyQualifiedName. If the 

    type is in the currently executing assembly or in Mscorlib.dll,it is sufficient to supply the 

    type name qualified by its namespace.

  

   throwOnError: true to throw an exception if the type cannot be found; false to return null. Specifying false 

    also suppresses some other exception conditions,but not all of them. See the Exceptions 

    section.

  

   Returns: The type with the specified name. If the type is not found,the throwOnError parameter specifies 

    whether null is returned or an exception is thrown. In some cases,an exception is thrown 

    regardless of the value of throwOnError. See the Exceptions section.

  

  GetType(typeName: str) -> Type

  

   Gets the System.Type with the specified name,performing a case-sensitive search.

  

   typeName: The assembly-qualified name of the type to get. See System.Type.AssemblyQualifiedName. If the 

    type is in the currently executing assembly or in Mscorlib.dll,it is sufficient to supply the 

    type name qualified by its namespace.

  

   Returns: The type with the specified name,if found; otherwise,null.
  """
  pass
 @staticmethod
 def GetTypeArray(args):
  """
  GetTypeArray(args: Array[object]) -> Array[Type]

  

   Gets the types of the objects in the specified array.

  

   args: An array of objects whose types to determine.

   Returns: An array of System.Type objects representing the types of the corresponding elements in args.
  """
  pass
 @staticmethod
 def GetTypeCode(type):
  """
  GetTypeCode(type: Type) -> TypeCode

  

   Gets the underlying type code of the specified System.Type.

  

   type: The type whose underlying type code to get.

   Returns: The code of the underlying type.
  """
  pass
 def GetTypeCodeImpl(self,*args):
  """
  GetTypeCodeImpl(self: Type) -> TypeCode

  

   Returns the underlying type code of the specified System.Type.

   Returns: The code of the underlying type.
  """
  pass
 @staticmethod
 def GetTypeFromCLSID(clsid,*__args):
  """
  GetTypeFromCLSID(clsid: Guid,server: str) -> Type

  

   Gets the type associated with the specified class identifier (CLSID) from the specified server.

  

   clsid: The CLSID of the type to get.

   server: The server from which to load the type. If the server name is null,this method automatically 

    reverts to the local machine.

  

   Returns: System.__ComObject regardless of whether the CLSID is valid.

  GetTypeFromCLSID(clsid: Guid,server: str,throwOnError: bool) -> Type

  

   Gets the type associated with the specified class identifier (CLSID) from the specified server,

    specifying whether to throw an exception if an error occurs while loading the type.

  

  

   clsid: The CLSID of the type to get.

   server: The server from which to load the type. If the server name is null,this method automatically 

    reverts to the local machine.

  

   throwOnError: true to throw any exception that occurs.-or- false to ignore any exception that occurs.

   Returns: System.__ComObject regardless of whether the CLSID is valid.

  GetTypeFromCLSID(clsid: Guid) -> Type

  

   Gets the type associated with the specified class identifier (CLSID).

  

   clsid: The CLSID of the type to get.

   Returns: System.__ComObject regardless of whether the CLSID is valid.

  GetTypeFromCLSID(clsid: Guid,throwOnError: bool) -> Type

  

   Gets the type associated with the specified class identifier (CLSID),specifying whether to 

    throw an exception if an error occurs while loading the type.

  

  

   clsid: The CLSID of the type to get.

   throwOnError: true to throw any exception that occurs.-or- false to ignore any exception that occurs.

   Returns: System.__ComObject regardless of whether the CLSID is valid.
  """
  pass
 @staticmethod
 def GetTypeFromHandle(handle):
  """
  GetTypeFromHandle(handle: RuntimeTypeHandle) -> Type

  

   Gets the type referenced by the specified type handle.

  

   handle: The object that refers to the type.

   Returns: The type referenced by the specified System.RuntimeTypeHandle,or null if the 

    System.RuntimeTypeHandle.Value property of handle is null.
  """
  pass
 @staticmethod
 def GetTypeFromProgID(progID,*__args):
  """
  GetTypeFromProgID(progID: str,server: str) -> Type

  

   Gets the type associated with the specified program identifier (progID) from the specified 

    server,returning null if an error is encountered while loading the type.

  

  

   progID: The progID of the type to get.

   server: The server from which to load the type. If the server name is null,this method automatically 

    reverts to the local machine.

  

   Returns: The type associated with the specified program identifier (progID),if progID is a valid entry 

    in the registry and a type is associated with it; otherwise,null.

  

  GetTypeFromProgID(progID: str,server: str,throwOnError: bool) -> Type

  

   Gets the type associated with the specified program identifier (progID) from the specified 

    server,specifying whether to throw an exception if an error occurs while loading the type.

  

  

   progID: The progID of the System.Type to get.

   server: The server from which to load the type. If the server name is null,this method automatically 

    reverts to the local machine.

  

   throwOnError: true to throw any exception that occurs.-or- false to ignore any exception that occurs.

   Returns: The type associated with the specified program identifier (progID),if progID is a valid entry 

    in the registry and a type is associated with it; otherwise,null.

  

  GetTypeFromProgID(progID: str) -> Type

  

   Gets the type associated with the specified program identifier (ProgID),returning null if an 

    error is encountered while loading the System.Type.

  

  

   progID: The ProgID of the type to get.

   Returns: The type associated with the specified ProgID,if progID is a valid entry in the registry and a 

    type is associated with it; otherwise,null.

  

  GetTypeFromProgID(progID: str,throwOnError: bool) -> Type

  

   Gets the type associated with the specified program identifier (ProgID),specifying whether to 

    throw an exception if an error occurs while loading the type.

  

  

   progID: The ProgID of the type to get.

   throwOnError: true to throw any exception that occurs.-or- false to ignore any exception that occurs.

   Returns: The type associated with the specified program identifier (ProgID),if progID is a valid entry 

    in the registry and a type is associated with it; otherwise,null.
  """
  pass
 @staticmethod
 def GetTypeHandle(o):
  """
  GetTypeHandle(o: object) -> RuntimeTypeHandle

  

   Gets the handle for the System.Type of a specified object.

  

   o: The object for which to get the type handle.

   Returns: The handle for the System.Type of the specified System.Object.
  """
  pass
 def HasElementTypeImpl(self,*args):
  """
  HasElementTypeImpl(self: Type) -> bool

  

   When overridden in a derived class,implements the System.Type.HasElementType property and 

    determines whether the current System.Type encompasses or refers to another type; that is,

    whether the current System.Type is an array,a pointer,or is passed by reference.

  

   Returns: true if the System.Type is an array,a pointer,or is passed by reference; otherwise,false.
  """
  pass
 def InvokeMember(self,name,invokeAttr,binder,target,args,*__args):
  """
  InvokeMember(self: Type,name: str,invokeAttr: BindingFlags,binder: Binder,target: object,args: Array[object]) -> object

  

   Invokes the specified member,using the specified binding constraints and matching the specified 

    argument list.

  

  

   name: The string containing the name of the constructor,method,property,or field member to 

    invoke.-or- An empty string ("") to invoke the default member. -or-For IDispatch members,a 

    string representing the DispID,for example "[DispID=3]".

  

   invokeAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted. The access can be one of the BindingFlags such as Public,NonPublic,Private,

    InvokeMethod,GetField,and so on. The type of lookup need not be specified. If the type of 

    lookup is omitted,BindingFlags.Public | BindingFlags.Instance | BindingFlags.Static are used.

  

   binder: An object that defines a set of properties and enables binding,which can involve selection of 

    an overloaded method,coercion of argument types,and invocation of a member through 

    reflection.-or- A null reference (Nothing in Visual Basic),to use the 

    System.Type.DefaultBinder. Note that explicitly defining a System.Reflection.Binder object may 

    be required for successfully invoking method overloads with variable arguments.

  

   target: The object on which to invoke the specified member.

   args: An array containing the arguments to pass to the member to invoke.

   Returns: An object representing the return value of the invoked member.

  InvokeMember(self: Type,name: str,invokeAttr: BindingFlags,binder: Binder,target: object,args: Array[object],culture: CultureInfo) -> object

  

   Invokes the specified member,using the specified binding constraints and matching the specified 

    argument list and culture.

  

  

   name: The string containing the name of the constructor,method,property,or field member to 

    invoke.-or- An empty string ("") to invoke the default member. -or-For IDispatch members,a 

    string representing the DispID,for example "[DispID=3]".

  

   invokeAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted. The access can be one of the BindingFlags such as Public,NonPublic,Private,

    InvokeMethod,GetField,and so on. The type of lookup need not be specified. If the type of 

    lookup is omitted,BindingFlags.Public | BindingFlags.Instance | BindingFlags.Static are used.

  

   binder: An object that defines a set of properties and enables binding,which can involve selection of 

    an overloaded method,coercion of argument types,and invocation of a member through 

    reflection.-or- A null reference (Nothing in Visual Basic),to use the 

    System.Type.DefaultBinder. Note that explicitly defining a System.Reflection.Binder object may 

    be required for successfully invoking method overloads with variable arguments.

  

   target: The object on which to invoke the specified member.

   args: An array containing the arguments to pass to the member to invoke.

   culture: The object representing the globalization locale to use,which may be necessary for 

    locale-specific conversions,such as converting a numeric System.String to a System.Double.-or- 

    A null reference (Nothing in Visual Basic) to use the current thread's 

    System.Globalization.CultureInfo.

  

   Returns: An object representing the return value of the invoked member.

  InvokeMember(self: Type,name: str,invokeAttr: BindingFlags,binder: Binder,target: object,args: Array[object],modifiers: Array[ParameterModifier],culture: CultureInfo,namedParameters: Array[str]) -> object

  

   When overridden in a derived class,invokes the specified member,using the specified binding 

    constraints and matching the specified argument list,modifiers and culture.

  

  

   name: The string containing the name of the constructor,method,property,or field member to 

    invoke.-or- An empty string ("") to invoke the default member. -or-For IDispatch members,a 

    string representing the DispID,for example "[DispID=3]".

  

   invokeAttr: A bitmask comprised of one or more System.Reflection.BindingFlags that specify how the search is 

    conducted. The access can be one of the BindingFlags such as Public,NonPublic,Private,

    InvokeMethod,GetField,and so on. The type of lookup need not be specified. If the type of 

    lookup is omitted,BindingFlags.Public | BindingFlags.Instance | BindingFlags.Static are used.

  

   binder: An object that defines a set of properties and enables binding,which can involve selection of 

    an overloaded method,coercion of argument types,and invocation of a member through 

    reflection.-or- A null reference (Nothing in Visual Basic),to use the 

    System.Type.DefaultBinder. Note that explicitly defining a System.Reflection.Binder object may 

    be required for successfully invoking method overloads with variable arguments.

  

   target: The object on which to invoke the specified member.

   args: An array containing the arguments to pass to the member to invoke.

   modifiers: An array of System.Reflection.ParameterModifier objects representing the attributes associated 

    with the corresponding element in the args array. A parameter's associated attributes are stored 

    in the member's signature. The default binder processes this parameter only when calling a COM 

    component.

  

   culture: The System.Globalization.CultureInfo object representing the globalization locale to use,which 

    may be necessary for locale-specific conversions,such as converting a numeric String to a 

    Double.-or- A null reference (Nothing in Visual Basic) to use the current thread's 

    System.Globalization.CultureInfo.

  

   namedParameters: An array containing the names of the parameters to which the values in the args array are passed.

   Returns: An object representing the return value of the invoked member.
  """
  pass
 def IsArrayImpl(self,*args):
  """
  IsArrayImpl(self: Type) -> bool

  

   When overridden in a derived class,implements the System.Type.IsArray property and determines 

    whether the System.Type is an array.

  

   Returns: true if the System.Type is an array; otherwise,false.
  """
  pass
 def IsAssignableFrom(self,c):
  """
  IsAssignableFrom(self: Type,c: Type) -> bool

  

   Determines whether an instance of the current System.Type can be assigned from an instance of 

    the specified Type.

  

  

   c: The type to compare with the current type.

   Returns: true if c and the current Type represent the same type,or if the current Type is in the 

    inheritance hierarchy of c,or if the current Type is an interface that c implements,or if c is 

    a generic type parameter and the current Type represents one of the constraints of c. false if 

    none of these conditions are true,or if c is null.
  """
  pass
 def IsByRefImpl(self,*args):
  """
  IsByRefImpl(self: Type) -> bool

  

   When overridden in a derived class,implements the System.Type.IsByRef property and determines 

    whether the System.Type is passed by reference.

  

   Returns: true if the System.Type is passed by reference; otherwise,false.
  """
  pass
 def IsCOMObjectImpl(self,*args):
  """
  IsCOMObjectImpl(self: Type) -> bool

  

   When overridden in a derived class,implements the System.Type.IsCOMObject property and 

    determines whether the System.Type is a COM object.

  

   Returns: true if the System.Type is a COM object; otherwise,false.
  """
  pass
 def IsContextfulImpl(self,*args):
  """
  IsContextfulImpl(self: Type) -> bool

  

   Implements the System.Type.IsContextful property and determines whether the System.Type can be 

    hosted in a context.

  

   Returns: true if the System.Type can be hosted in a context; otherwise,false.
  """
  pass
 def IsEnumDefined(self,value):
  """
  IsEnumDefined(self: Type,value: object) -> bool

  

   Returns a value that indicates whether the specified value exists in the current enumeration 

    type.

  

  

   value: The value to be tested.

   Returns: true if the specified value is a member of the current enumeration type; otherwise,false.
  """
  pass
 def IsEquivalentTo(self,other):
  """
  IsEquivalentTo(self: Type,other: Type) -> bool

  

   Determines whether two COM types have the same identity and are eligible for type equivalence.

  

   other: The COM type that is tested for equivalence with the current type.

   Returns: true if the COM types are equivalent; otherwise,false. This method also returns false if one 

    type is in an assembly that is loaded for execution,and the other is in an assembly that is 

    loaded into the reflection-only context.
  """
  pass
 def IsInstanceOfType(self,o):
  """
  IsInstanceOfType(self: Type,o: object) -> bool

  

   Determines whether the specified object is an instance of the current System.Type.

  

   o: The object to compare with the current type.

   Returns: true if the current Type is in the inheritance hierarchy of the object represented by o,or if 

    the current Type is an interface that o supports. false if neither of these conditions is the 

    case,or if o is null,or if the current Type is an open generic type (that is,

    System.Type.ContainsGenericParameters returns true).
  """
  pass
 def IsMarshalByRefImpl(self,*args):
  """
  IsMarshalByRefImpl(self: Type) -> bool

  

   Implements the System.Type.IsMarshalByRef property and determines whether the System.Type is 

    marshaled by reference.

  

   Returns: true if the System.Type is marshaled by reference; otherwise,false.
  """
  pass
 def IsPointerImpl(self,*args):
  """
  IsPointerImpl(self: Type) -> bool

  

   When overridden in a derived class,implements the System.Type.IsPointer property and determines 

    whether the System.Type is a pointer.

  

   Returns: true if the System.Type is a pointer; otherwise,false.
  """
  pass
 def IsPrimitiveImpl(self,*args):
  """
  IsPrimitiveImpl(self: Type) -> bool

  

   When overridden in a derived class,implements the System.Type.IsPrimitive property and 

    determines whether the System.Type is one of the primitive types.

  

   Returns: true if the System.Type is one of the primitive types; otherwise,false.
  """
  pass
 def IsSubclassOf(self,c):
  """
  IsSubclassOf(self: Type,c: Type) -> bool

  

   Determines whether the class represented by the current System.Type derives from the class 

    represented by the specified System.Type.

  

  

   c: The type to compare with the current type.

   Returns: true if the Type represented by the c parameter and the current Type represent classes,and the 

    class represented by the current Type derives from the class represented by c; otherwise,false. 

    This method also returns false if c and the current Type represent the same class.
  """
  pass
 def IsValueTypeImpl(self,*args):
  """
  IsValueTypeImpl(self: Type) -> bool

  

   Implements the System.Type.IsValueType property and determines whether the System.Type is a 

    value type; that is,not a class or an interface.

  

   Returns: true if the System.Type is a value type; otherwise,false.
  """
  pass
 def MakeArrayType(self,rank=None):
  """
  MakeArrayType(self: Type,rank: int) -> Type

  

   Returns a System.Type object representing an array of the current type,with the specified 

    number of dimensions.

  

  

   rank: The number of dimensions for the array. This number must be less than or equal to 32.

   Returns: An object representing an array of the current type,with the specified number of dimensions.

  MakeArrayType(self: Type) -> Type

  

   Returns a System.Type object representing a one-dimensional array of the current type,with a 

    lower bound of zero.

  

   Returns: A System.Type object representing a one-dimensional array of the current type,with a lower 

    bound of zero.
  """
  pass
 def MakeByRefType(self):
  """
  MakeByRefType(self: Type) -> Type

  

   Returns a System.Type object that represents the current type when passed as a ref parameter 

    (ByRef parameter in Visual Basic).

  

   Returns: A System.Type object that represents the current type when passed as a ref parameter (ByRef 

    parameter in Visual Basic).
  """
  pass
 def MakeGenericType(self,typeArguments):
  """
  MakeGenericType(self: Type,*typeArguments: Array[Type]) -> Type

  

   Substitutes the elements of an array of types for the type parameters of the current generic 

    type definition and returns a System.Type object representing the resulting constructed type.

  

  

   typeArguments: An array of types to be substituted for the type parameters of the current generic type.

   Returns: A System.Type representing the constructed type formed by substituting the elements of 

    typeArguments for the type parameters of the current generic type.
  """
  pass
 def MakePointerType(self):
  """
  MakePointerType(self: Type) -> Type

  

   Returns a System.Type object that represents a pointer to the current type.

   Returns: A System.Type object that represents a pointer to the current type.
  """
  pass
 @staticmethod
 def ReflectionOnlyGetType(typeName,throwIfNotFound,ignoreCase):
  """
  ReflectionOnlyGetType(typeName: str,throwIfNotFound: bool,ignoreCase: bool) -> Type

  

   Gets the System.Type with the specified name,specifying whether to perform a case-sensitive 

    search and whether to throw an exception if the type is not found. The type is loaded for 

    reflection only,not for execution.

  

  

   typeName: The assembly-qualified name of the System.Type to get.

   throwIfNotFound: true to throw a System.TypeLoadException if the type cannot be found; false to return null if 

    the type cannot be found. Specifying false also suppresses some other exception conditions,but 

    not all of them. See the Exceptions section.

  

   ignoreCase: true to perform a case-insensitive search for typeName; false to perform a case-sensitive search 

    for typeName.

  

   Returns: The type with the specified name,if found; otherwise,null. If the type is not found,the 

    throwIfNotFound parameter specifies whether null is returned or an exception is thrown. In some 

    cases,an exception is thrown regardless of the value of throwIfNotFound. See the Exceptions 

    section.
  """
  pass
 def ToString(self):
  """
  ToString(self: Type) -> str

  

   Returns a String representing the name of the current Type.

   Returns: A System.String representing the name of the current System.Type.
  """
  pass
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==y """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 Assembly=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the System.Reflection.Assembly in which the type is declared. For generic types,gets the System.Reflection.Assembly in which the generic type is defined.



Get: Assembly(self: Type) -> Assembly



"""

 AssemblyQualifiedName=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the assembly-qualified name of the System.Type,which includes the name of the assembly from which the System.Type was loaded.



Get: AssemblyQualifiedName(self: Type) -> str



"""

 Attributes=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the attributes associated with the System.Type.



Get: Attributes(self: Type) -> TypeAttributes



"""

 BaseType=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the type from which the current System.Type directly inherits.



Get: BaseType(self: Type) -> Type



"""

 ContainsGenericParameters=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the current System.Type object has type parameters that have not been replaced by specific types.



Get: ContainsGenericParameters(self: Type) -> bool



"""

 DeclaringMethod=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a System.Reflection.MethodBase that represents the declaring method,if the current System.Type represents a type parameter of a generic method.



Get: DeclaringMethod(self: Type) -> MethodBase



"""

 DeclaringType=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the type that declares the current nested type or generic type parameter.



Get: DeclaringType(self: Type) -> Type



"""

 FullName=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the fully qualified name of the System.Type,including the namespace of the System.Type but not the assembly.



Get: FullName(self: Type) -> str



"""

 GenericParameterAttributes=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a combination of System.Reflection.GenericParameterAttributes flags that describe the covariance and special constraints of the current generic type parameter.



Get: GenericParameterAttributes(self: Type) -> GenericParameterAttributes



"""

 GenericParameterPosition=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the position of the type parameter in the type parameter list of the generic type or method that declared the parameter,when the System.Type object represents a type parameter of a generic type or a generic method.



Get: GenericParameterPosition(self: Type) -> int



"""

 GenericTypeArguments=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Get: GenericTypeArguments(self: Type) -> Array[Type]



"""

 GUID=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the GUID associated with the System.Type.



Get: GUID(self: Type) -> Guid



"""

 HasElementType=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the current System.Type encompasses or refers to another type; that is,whether the current System.Type is an array,a pointer,or is passed by reference.



Get: HasElementType(self: Type) -> bool



"""

 IsAbstract=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type is abstract and must be overridden.



Get: IsAbstract(self: Type) -> bool



"""

 IsAnsiClass=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the string format attribute AnsiClass is selected for the System.Type.



Get: IsAnsiClass(self: Type) -> bool



"""

 IsArray=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type is an array.



Get: IsArray(self: Type) -> bool



"""

 IsAutoClass=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the string format attribute AutoClass is selected for the System.Type.



Get: IsAutoClass(self: Type) -> bool



"""

 IsAutoLayout=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the class layout attribute AutoLayout is selected for the System.Type.



Get: IsAutoLayout(self: Type) -> bool



"""

 IsByRef=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type is passed by reference.



Get: IsByRef(self: Type) -> bool



"""

 IsClass=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type is a class; that is,not a value type or interface.



Get: IsClass(self: Type) -> bool



"""

 IsCOMObject=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type is a COM object.



Get: IsCOMObject(self: Type) -> bool



"""

 IsConstructedGenericType=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Get: IsConstructedGenericType(self: Type) -> bool



"""

 IsContextful=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type can be hosted in a context.



Get: IsContextful(self: Type) -> bool



"""

 IsEnum=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the current System.Type represents an enumeration.



Get: IsEnum(self: Type) -> bool



"""

 IsExplicitLayout=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the class layout attribute ExplicitLayout is selected for the System.Type.



Get: IsExplicitLayout(self: Type) -> bool



"""

 IsGenericParameter=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the current System.Type represents a type parameter in the definition of a generic type or method.



Get: IsGenericParameter(self: Type) -> bool



"""

 IsGenericType=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the current type is a generic type.



Get: IsGenericType(self: Type) -> bool



"""

 IsGenericTypeDefinition=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the current System.Type represents a generic type definition,from which other generic types can be constructed.



Get: IsGenericTypeDefinition(self: Type) -> bool



"""

 IsImport=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type has a System.Runtime.InteropServices.ComImportAttribute attribute applied,indicating that it was imported from a COM type library.



Get: IsImport(self: Type) -> bool



"""

 IsInterface=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type is an interface; that is,not a class or a value type.



Get: IsInterface(self: Type) -> bool



"""

 IsLayoutSequential=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the class layout attribute SequentialLayout is selected for the System.Type.



Get: IsLayoutSequential(self: Type) -> bool



"""

 IsMarshalByRef=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type is marshaled by reference.



Get: IsMarshalByRef(self: Type) -> bool



"""

 IsNested=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the current System.Type object represents a type whose definition is nested inside the definition of another type.



Get: IsNested(self: Type) -> bool



"""

 IsNestedAssembly=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type is nested and visible only within its own assembly.



Get: IsNestedAssembly(self: Type) -> bool



"""

 IsNestedFamANDAssem=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type is nested and visible only to classes that belong to both its own family and its own assembly.



Get: IsNestedFamANDAssem(self: Type) -> bool



"""

 IsNestedFamily=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type is nested and visible only within its own family.



Get: IsNestedFamily(self: Type) -> bool



"""

 IsNestedFamORAssem=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type is nested and visible only to classes that belong to either its own family or to its own assembly.



Get: IsNestedFamORAssem(self: Type) -> bool



"""

 IsNestedPrivate=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type is nested and declared private.



Get: IsNestedPrivate(self: Type) -> bool



"""

 IsNestedPublic=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether a class is nested and declared public.



Get: IsNestedPublic(self: Type) -> bool



"""

 IsNotPublic=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type is not declared public.



Get: IsNotPublic(self: Type) -> bool



"""

 IsPointer=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type is a pointer.



Get: IsPointer(self: Type) -> bool



"""

 IsPrimitive=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type is one of the primitive types.



Get: IsPrimitive(self: Type) -> bool



"""

 IsPublic=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type is declared public.



Get: IsPublic(self: Type) -> bool



"""

 IsSealed=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type is declared sealed.



Get: IsSealed(self: Type) -> bool



"""

 IsSecurityCritical=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value that indicates whether the current type is security-critical or security-safe-critical at the current trust level,and therefore can perform critical operations.



Get: IsSecurityCritical(self: Type) -> bool



"""

 IsSecuritySafeCritical=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value that indicates whether the current type is security-safe-critical at the current trust level; that is,whether it can perform critical operations and can be accessed by transparent code.



Get: IsSecuritySafeCritical(self: Type) -> bool



"""

 IsSecurityTransparent=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value that indicates whether the current type is transparent at the current trust level,and therefore cannot perform critical operations.



Get: IsSecurityTransparent(self: Type) -> bool



"""

 IsSerializable=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type is serializable.



Get: IsSerializable(self: Type) -> bool



"""

 IsSpecialName=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type has a name that requires special handling.



Get: IsSpecialName(self: Type) -> bool



"""

 IsUnicodeClass=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the string format attribute UnicodeClass is selected for the System.Type.



Get: IsUnicodeClass(self: Type) -> bool



"""

 IsValueType=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type is a value type.



Get: IsValueType(self: Type) -> bool



"""

 IsVisible=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the System.Type can be accessed by code outside the assembly.



Get: IsVisible(self: Type) -> bool



"""

 MemberType=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a System.Reflection.MemberTypes value indicating that this member is a type or a nested type.



Get: MemberType(self: Type) -> MemberTypes



"""

 Module=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the module (the DLL) in which the current System.Type is defined.



Get: Module(self: Type) -> Module



"""

 Namespace=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the namespace of the System.Type.



Get: Namespace(self: Type) -> str



"""

 ReflectedType=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the class object that was used to obtain this member.



Get: ReflectedType(self: Type) -> Type



"""

 StructLayoutAttribute=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a System.Runtime.InteropServices.StructLayoutAttribute that describes the layout of the current type.



Get: StructLayoutAttribute(self: Type) -> StructLayoutAttribute



"""

 TypeHandle=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the handle for the current System.Type.



Get: TypeHandle(self: Type) -> RuntimeTypeHandle



"""

 TypeInitializer=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the initializer for the System.Type.



Get: TypeInitializer(self: Type) -> ConstructorInfo



"""

 UnderlyingSystemType=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Indicates the type provided by the common language runtime that represents this type.



Get: UnderlyingSystemType(self: Type) -> Type



"""


 DefaultBinder=None
 Delimiter=None
 EmptyTypes=None
 Missing=None




class TreeViewCancelEventArgs(CancelEventArgs):
 """
 Provides data for the System.Windows.Forms.TreeView.BeforeCheck,System.Windows.Forms.TreeView.BeforeCollapse,System.Windows.Forms.TreeView.BeforeExpand,and System.Windows.Forms.TreeView.BeforeSelect events of a System.Windows.Forms.TreeView control.

 

 TreeViewCancelEventArgs(node: TreeNode,cancel: bool,action: TreeViewAction)
 """
 @staticmethod
 def __new__(self,node,cancel,action):
  """ __new__(cls: type,node: TreeNode,cancel: bool,action: TreeViewAction) """
  pass
 Action=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the type of System.Windows.Forms.TreeView action that raised the event.



Get: Action(self: TreeViewCancelEventArgs) -> TreeViewAction



"""

 Node=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the tree node to be checked,expanded,collapsed,or selected.



Get: Node(self: TreeViewCancelEventArgs) -> TreeNode



"""





class TypeCode(Enum,IComparable,IFormattable,IConvertible):
 """
 Specifies the type of an object.

 

 enum TypeCode,values: Boolean (3),Byte (6),Char (4),DateTime (16),DBNull (2),Decimal (15),Double (14),Empty (0),Int16 (7),Int32 (9),Int64 (11),Object (1),SByte (5),Single (13),String (18),UInt16 (8),UInt32 (10),UInt64 (12)
 """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 Boolean=None
 Byte=None
 Char=None
 DateTime=None
 DBNull=None
 Decimal=None
 Double=None
 Empty=None
 Int16=None
 Int32=None
 Int64=None
 Object=None
 SByte=None
 Single=None
 String=None
 UInt16=None
 UInt32=None
 UInt64=None
 value__=None




"""
Wall Wrappers

"""  #

import rpw
from rpw import revit, DB
from rpw.db import Element
from rpw.db import FamilyInstance, FamilySymbol, Family, Category
from rpw.base import BaseObjectWrapper
from rpw.utils.logger import logger, deprecate_warning
from rpw.utils.coerce import to_element_id
from rpw.db.builtins import BipEnum
from rpw.exceptions import RpwTypeError, RpwCoerceError
from rpw.utils.mixins import ByNameCollectMixin


class Wall(FamilyInstance):
    """
    Inherits base ``FamilyInstance`` and overrides symbol attribute to
    get `Symbol` equivalent of Wall - WallType `(GetTypeId)`
    """

    _revit_object_category = DB.BuiltInCategory.OST_Walls
    _revit_object_class = DB.Wall
    _collector_params = {'of_class': _revit_object_class, 'is_type': False}

    def change_type(self, wall_type_reference):
        """
        Change Wall Type

        Args:
            wall_type_reference (``ElementId``, ``WallType``, ``str``): Wall Type Reference
        """
        wall_type = WallType.by_name_or_element_ref(wall_type_reference)
        wall_type_id = to_element_id(wall_type)
        self._revit_object.ChangeTypeId(wall_type_id)

    def get_symbol(self, wrapped=True):
        """ Get Wall Type Alias """
        return self.get_wall_type(wrapped)

    @property
    def symbol(self):
        deprecate_warning('Wall.symbol', 'Wall.get_symbol()')
        return self.get_symbol(wrapped=True)

    def get_wall_type(self, wrapped=True):
        """ Get Wall Type """
        wall_type_id = self._revit_object.GetTypeId()
        wall_type = self.doc.GetElement(wall_type_id)
        return WallType(wall_type) if wrapped else wall_type

    @property
    def wall_type(self):
        deprecate_warning('Wall.wall_type', 'Wall.get_wall_type()')
        return self.get_wall_type(wrapped=True)

    def get_wall_kind(self, wrapped=True):
        wall_type = self.get_wall_type(wrapped=True)
        return wall_type.get_wall_kind(wrapped=wrapped)

    @property
    def wall_kind(self):
        deprecate_warning('Wall.wall_kind', 'Wall.get_wall_kind()')
        return self.get_wall_kind(wrapped=True)

    def get_family(self, wrapped=True):
        """ Get WallKind Alias """
        return self.get_wall_kind(wrapped=wrapped)

    @property
    def family(self):
        deprecate_warning('Wall.family', 'Wall.get_family()')
        return self.get_family(wrapped=True)

    def get_category(self, wrapped=True):
        """ Get Wall Category """
        return WallCategory(self._revit_object.Category)

    @property
    def category(self):
        """ Wrapped ``DB.Category`` of the ``DB.Wall`` """
        deprecate_warning('Wall.category', 'Wall.get_category()')
        return self.get_category(wrapped=True)


class WallType(FamilySymbol, ByNameCollectMixin):
    """
    Inherits from :any:`FamilySymbol` and overrides:
        * :func:`wall_kind` to get the `Family` equivalent of Wall `(.Kind)`
        * Uses a different method to get instances.
    """

    _revit_object_class = DB.WallType
    _collector_params = {'of_class': _revit_object_class, 'is_type': True}

    def get_family(self, wrapped=True):
        return self.get_wall_kind(wrapped=wrapped)

    @property
    def family(self):
        deprecate_warning('WallType.family', 'WallType.get_family()')
        return self.get_wall_kind(wrapped=True)

    def get_wall_kind(self, wrapped=True):
        """ Returns ``DB.Family`` of the Symbol """
        kind = self._revit_object.Kind
        return WallKind(kind) if wrapped else kind

    @property
    def wall_kind(self):
        """ Returns ``DB.Family`` of the Symbol """
        deprecate_warning('WallType.wall_kind', 'WallType.get_wall_kind()')
        return self.get_wall_kind(wrapped=True)

    def get_instances(self, wrapped=True):
        """ Returns all Instances of this Wall Types """
        bip = BipEnum.get_id('SYMBOL_NAME_PARAM')
        param_filter = rpw.db.ParameterFilter(bip, equals=self.name)
        return rpw.db.Collector(parameter_filter=param_filter,
                                **Wall._collector_params).wrapped_elements

    @property
    def instances(self):
        """ Returns all Instances of this Wall Types """
        deprecate_warning('WallType.instances', 'WallType.get_instances()')
        return self.get_instances(wrapped=True)

    def get_siblings(self, wrapped=True):
        wall_kind = self.get_wall_kind(wrapped=True)
        return wall_kind.get_wall_types(wrapped=wrapped)

    @property
    def siblings(self):
        deprecate_warning('WallType.siblings', 'WallType.get_siblings()')
        return self.get_siblings(wrapped=True)

    def get_category(self, wrapped=True):
        """ Get Wall Category """
        return WallCategory(self._revit_object.Category)

    @property
    def category(self):
        """ Wrapped ``DB.Category`` of the ``DB.Wall`` """
        deprecate_warning('Wall.category', 'Wall.get_category()')
        return self.get_category(wrapped=True)


# class WallKind(Family):
class WallKind(BaseObjectWrapper):
    """
    Equivalent of ``Family`` but is Enumerator for Walls.

    Can be Basic, Stacked, Curtain, Unknown
    """

    _revit_object_class = DB.WallKind

    @property
    def name(self):
        """ Retuns Pretty Name as shown on UI: Basic > Basic Wall """
        # Same method as Family Works, but requires Code duplication
        # Since this should not inherit from Family.
        # Solution copy code or Mixin. Or return Enum Name:  'Basic'
        # This works but requires Lookup.
        # wall_type = self.get_wall_types()[0]
        # return wall_type.parameters.builtins['SYMBOL_FAMILY_NAME_PARAM'].value
        # return '{} Wall'.format(self._revit_object.ToString())
        return self._revit_object.ToString()

    def get_symbols(self, wrapped=True):
        """ Get Wall Types Alias """
        return self.get_wall_types(wrapped=wrapped)

    @property
    def symbols(self):
        deprecate_warning('WallKind.symbols', 'WallKind.get_symbols()')
        return self.get_symbols(wrapped=True)

    def get_wall_types(self, wrapped=True):
        """ Get Wall Types Alias """
        type_collector = rpw.db.WallType.collect()
        wall_types = type_collector.get_elements(wrapped=wrapped)
        return [wall_type for wall_type in wall_types
                if wall_type.Kind == self._revit_object]

    @property
    def wall_types(self):
        deprecate_warning('WallKind.wall_types', 'WallKind.get_wall_types()')
        return self.get_wall_types(wrapped=True)

    def get_instances(self, wrapped=True):
        """ Returns all Wall instances of this given Wall Kind"""
        instances = []
        for wall_type in self.get_wall_types(wrapped=True):
            instances.extend(wall_type.get_instances(wrapped=wrapped))
        return instances

    @property
    def instances(self):
        """ Returns all Wall instances of this given Wall Kind"""
        deprecate_warning('WallKind.instances', 'WallKind.get_instances()')
        return self.get_instances(wrapped=True)

    def get_category(self, wrapped=True):
        cat = DB.Category.GetCategory(revit.doc, DB.BuiltInCategory.OST_Walls)
        return WallCategory(cat) if wrapped else cat

    @property
    def category(self):
        deprecate_warning('WallKind.category', 'WallKind.get_category()')
        return self.get_category(wrapped=True)

    def __repr__(self):
        return super(WallKind, self).__repr__({'name': self.name})

class WallCategory(Category):
    """
    ``DB.Category`` Wall Category Wrapper

    Attribute:
        _revit_object (DB.Family): Wrapped ``DB.Category``
    """

    _revit_object_class = DB.Category

    def get_families(self, wrapped=True):
        """ Returns ``DB.WallKind`` elements in the category """
        wall_kinds = []
        for member in dir(DB.WallKind):
            if type(getattr(DB.WallKind, member)) is DB.WallKind:
                wall_kind = getattr(DB.WallKind, member)
                wall_kind = WallKind(wall_kind) if wrapped else wall_kind
                wall_kinds.append(wall_kind)
        return wall_kinds

    @property
    def families(self):
        """ Returns ``DB.WallKind`` elements in the category """
        deprecate_warning('WallCategory.families',
                          'WallCategory.get_families()')
        return self.get_families(wrapped=True)



class ToolStripRenderer(object):
 """ Handles the painting functionality for System.Windows.Forms.ToolStrip objects. """
 @staticmethod
 def CreateDisabledImage(normalImage):
  """
  CreateDisabledImage(normalImage: Image) -> Image

  

   Creates a gray-scale copy of a given image.

  

   normalImage: The image to be copied.

   Returns: An System.Drawing.Image that is a copy of the given image,but with a gray-scale color matrix.
  """
  pass
 def DrawArrow(self,e):
  """
  DrawArrow(self: ToolStripRenderer,e: ToolStripArrowRenderEventArgs)

   Draws an arrow on a System.Windows.Forms.ToolStripItem.

  

   e: A System.Windows.Forms.ToolStripArrowRenderEventArgs that contains data to draw the arrow.
  """
  pass
 def DrawButtonBackground(self,e):
  """
  DrawButtonBackground(self: ToolStripRenderer,e: ToolStripItemRenderEventArgs)

   Draws the background for a System.Windows.Forms.ToolStripButton.

  

   e: The System.Windows.Forms.ToolStripItemRenderEventArgs that contains data to draw the button's 

    background.
  """
  pass
 def DrawDropDownButtonBackground(self,e):
  """
  DrawDropDownButtonBackground(self: ToolStripRenderer,e: ToolStripItemRenderEventArgs)

   Draws the background for a System.Windows.Forms.ToolStripDropDownButton.

  

   e: A System.Windows.Forms.ToolStripItemRenderEventArgs that contains the data to draw the drop-down 

    button's background.
  """
  pass
 def DrawGrip(self,e):
  """
  DrawGrip(self: ToolStripRenderer,e: ToolStripGripRenderEventArgs)

   Draws a move handle on a System.Windows.Forms.ToolStrip.

  

   e: A System.Windows.Forms.ToolStripGripRenderEventArgs that contains the data to draw the move 

    handle.
  """
  pass
 def DrawImageMargin(self,e):
  """
  DrawImageMargin(self: ToolStripRenderer,e: ToolStripRenderEventArgs)

   Draws the space around an image on a System.Windows.Forms.ToolStrip.

  

   e: A System.Windows.Forms.ToolStripRenderEventArgs that contains the data to draw the space around 

    the image.
  """
  pass
 def DrawItemBackground(self,e):
  """
  DrawItemBackground(self: ToolStripRenderer,e: ToolStripItemRenderEventArgs)

   Draws the background for a System.Windows.Forms.ToolStripItem.

  

   e: A System.Windows.Forms.ToolStripItemRenderEventArgs that contains the data to draw the 

    background of the item.
  """
  pass
 def DrawItemCheck(self,e):
  """
  DrawItemCheck(self: ToolStripRenderer,e: ToolStripItemImageRenderEventArgs)

   Draws an image on a System.Windows.Forms.ToolStripItem that indicates the item is in a selected 

    state.

  

  

   e: A System.Windows.Forms.ToolStripItemImageRenderEventArgs that contains the data to draw the 

    selected image.
  """
  pass
 def DrawItemImage(self,e):
  """
  DrawItemImage(self: ToolStripRenderer,e: ToolStripItemImageRenderEventArgs)

   Draws an image on a System.Windows.Forms.ToolStripItem.

  

   e: A System.Windows.Forms.ToolStripItemImageRenderEventArgs that contains the data to draw the 

    image.
  """
  pass
 def DrawItemText(self,e):
  """
  DrawItemText(self: ToolStripRenderer,e: ToolStripItemTextRenderEventArgs)

   Draws text on a System.Windows.Forms.ToolStripItem.

  

   e: A System.Windows.Forms.ToolStripItemTextRenderEventArgs that contains the data to draw the text.
  """
  pass
 def DrawLabelBackground(self,e):
  """
  DrawLabelBackground(self: ToolStripRenderer,e: ToolStripItemRenderEventArgs)

   Draws the background for a System.Windows.Forms.ToolStripLabel.

  

   e: A System.Windows.Forms.ToolStripItemRenderEventArgs that contains the data to draw the 

    background for the label.
  """
  pass
 def DrawMenuItemBackground(self,e):
  """
  DrawMenuItemBackground(self: ToolStripRenderer,e: ToolStripItemRenderEventArgs)

   Draws the background for a System.Windows.Forms.ToolStripMenuItem.

  

   e: A System.Windows.Forms.ToolStripItemRenderEventArgs that contains the data to draw the 

    background for the menu item.
  """
  pass
 def DrawOverflowButtonBackground(self,e):
  """
  DrawOverflowButtonBackground(self: ToolStripRenderer,e: ToolStripItemRenderEventArgs)

   Draws the background for an overflow button.

  

   e: A System.Windows.Forms.ToolStripItemRenderEventArgs that contains the event data.
  """
  pass
 def DrawSeparator(self,e):
  """
  DrawSeparator(self: ToolStripRenderer,e: ToolStripSeparatorRenderEventArgs)

   Draws a System.Windows.Forms.ToolStripSeparator.

  

   e: A System.Windows.Forms.ToolStripSeparatorRenderEventArgs that contains the data to draw the 

    System.Windows.Forms.ToolStripSeparator.
  """
  pass
 def DrawSplitButton(self,e):
  """
  DrawSplitButton(self: ToolStripRenderer,e: ToolStripItemRenderEventArgs)

   Draws a System.Windows.Forms.ToolStripSplitButton.

  

   e: A System.Windows.Forms.ToolStripItemRenderEventArgs that contains the event data.
  """
  pass
 def DrawStatusStripSizingGrip(self,e):
  """
  DrawStatusStripSizingGrip(self: ToolStripRenderer,e: ToolStripRenderEventArgs)

   Draws a sizing grip.

  

   e: A System.Windows.Forms.ToolStripRenderEventArgs that contains the event data.
  """
  pass
 def DrawToolStripBackground(self,e):
  """
  DrawToolStripBackground(self: ToolStripRenderer,e: ToolStripRenderEventArgs)

   Draws the background for a System.Windows.Forms.ToolStrip.

  

   e: A System.Windows.Forms.ToolStripRenderEventArgs that contains the data to draw the background 

    for the System.Windows.Forms.ToolStrip.
  """
  pass
 def DrawToolStripBorder(self,e):
  """
  DrawToolStripBorder(self: ToolStripRenderer,e: ToolStripRenderEventArgs)

   Draws the border for a System.Windows.Forms.ToolStrip.

  

   e: A System.Windows.Forms.ToolStripRenderEventArgs that contains the data to draw the border for 

    the System.Windows.Forms.ToolStrip.
  """
  pass
 def DrawToolStripContentPanelBackground(self,e):
  """
  DrawToolStripContentPanelBackground(self: ToolStripRenderer,e: ToolStripContentPanelRenderEventArgs)

   Draws the background of the System.Windows.Forms.ToolStripContentPanel.

  

   e: A System.Windows.Forms.ToolStripContentPanelRenderEventArgs that contains the event data.
  """
  pass
 def DrawToolStripPanelBackground(self,e):
  """
  DrawToolStripPanelBackground(self: ToolStripRenderer,e: ToolStripPanelRenderEventArgs)

   Draws the background of the System.Windows.Forms.ToolStripPanel.

  

   e: A System.Windows.Forms.ToolStripPanelRenderEventArgs that contains the event data.
  """
  pass
 def DrawToolStripStatusLabelBackground(self,e):
  """
  DrawToolStripStatusLabelBackground(self: ToolStripRenderer,e: ToolStripItemRenderEventArgs)

   Draws the background of the System.Windows.Forms.ToolStripStatusLabel.

  

   e: A System.Windows.Forms.ToolStripItemRenderEventArgs that contains the event data.
  """
  pass
 def Initialize(self,*args):
  """
  Initialize(self: ToolStripRenderer,toolStrip: ToolStrip)

   When overridden in a derived class,provides for custom initialization of the given 

    System.Windows.Forms.ToolStrip.

  

  

   toolStrip: The System.Windows.Forms.ToolStrip to be initialized.
  """
  pass
 def InitializeContentPanel(self,*args):
  """
  InitializeContentPanel(self: ToolStripRenderer,contentPanel: ToolStripContentPanel)

   Initializes the specified System.Windows.Forms.ToolStripContentPanel.

  

   contentPanel: The System.Windows.Forms.ToolStripContentPanel.
  """
  pass
 def InitializeItem(self,*args):
  """
  InitializeItem(self: ToolStripRenderer,item: ToolStripItem)

   When overridden in a derived class,provides for custom initialization of the given 

    System.Windows.Forms.ToolStripItem.

  

  

   item: The System.Windows.Forms.ToolStripItem to be initialized.
  """
  pass
 def InitializePanel(self,*args):
  """
  InitializePanel(self: ToolStripRenderer,toolStripPanel: ToolStripPanel)

   Initializes the specified System.Windows.Forms.ToolStripPanel.

  

   toolStripPanel: The System.Windows.Forms.ToolStripPanel.
  """
  pass
 def OnRenderArrow(self,*args):
  """
  OnRenderArrow(self: ToolStripRenderer,e: ToolStripArrowRenderEventArgs)

   Raises the System.Windows.Forms.ToolStripRenderer.RenderArrow event.

  

   e: A System.Windows.Forms.ToolStripArrowRenderEventArgs that contains the event data.
  """
  pass
 def OnRenderButtonBackground(self,*args):
  """
  OnRenderButtonBackground(self: ToolStripRenderer,e: ToolStripItemRenderEventArgs)

   Raises the System.Windows.Forms.ToolStripRenderer.RenderButtonBackground event.

  

   e: A System.Windows.Forms.ToolStripRenderEventArgs that contains the event data.
  """
  pass
 def OnRenderDropDownButtonBackground(self,*args):
  """
  OnRenderDropDownButtonBackground(self: ToolStripRenderer,e: ToolStripItemRenderEventArgs)

   Raises the System.Windows.Forms.ToolStripRenderer.RenderDropDownButtonBackground event.

  

   e: A System.Windows.Forms.ToolStripItemRenderEventArgs that contains the event data.
  """
  pass
 def OnRenderGrip(self,*args):
  """
  OnRenderGrip(self: ToolStripRenderer,e: ToolStripGripRenderEventArgs)

   Raises the System.Windows.Forms.ToolStripRenderer.RenderGrip event.

  

   e: A System.Windows.Forms.ToolStripGripRenderEventArgs that contains the event data.
  """
  pass
 def OnRenderImageMargin(self,*args):
  """
  OnRenderImageMargin(self: ToolStripRenderer,e: ToolStripRenderEventArgs)

   Draws the item background.

  

   e: A System.Windows.Forms.ToolStripRenderEventArgs that contains the event data.
  """
  pass
 def OnRenderItemBackground(self,*args):
  """
  OnRenderItemBackground(self: ToolStripRenderer,e: ToolStripItemRenderEventArgs)

   Raises the 

    System.Windows.Forms.ToolStripSystemRenderer.OnRenderItemBackground(System.Windows.Forms.ToolStri

    pItemRenderEventArgs) event.

  

  

   e: A System.Windows.Forms.ToolStripItemRenderEventArgs that contains the event data.
  """
  pass
 def OnRenderItemCheck(self,*args):
  """
  OnRenderItemCheck(self: ToolStripRenderer,e: ToolStripItemImageRenderEventArgs)

   Raises the System.Windows.Forms.ToolStripRenderer.RenderItemCheck event.

  

   e: A System.Windows.Forms.ToolStripItemImageRenderEventArgs that contains the event data.
  """
  pass
 def OnRenderItemImage(self,*args):
  """
  OnRenderItemImage(self: ToolStripRenderer,e: ToolStripItemImageRenderEventArgs)

   Raises the System.Windows.Forms.ToolStripRenderer.RenderItemImage event.

  

   e: A System.Windows.Forms.ToolStripItemImageRenderEventArgs that contains the event data.
  """
  pass
 def OnRenderItemText(self,*args):
  """
  OnRenderItemText(self: ToolStripRenderer,e: ToolStripItemTextRenderEventArgs)

   Raises the System.Windows.Forms.ToolStripRenderer.RenderItemText event.

  

   e: A System.Windows.Forms.ToolStripItemTextRenderEventArgs that contains the event data.
  """
  pass
 def OnRenderLabelBackground(self,*args):
  """
  OnRenderLabelBackground(self: ToolStripRenderer,e: ToolStripItemRenderEventArgs)

   Raises the System.Windows.Forms.ToolStripRenderer.RenderLabelBackground event.

  

   e: A System.Windows.Forms.ToolStripItemRenderEventArgs that contains the event data.
  """
  pass
 def OnRenderMenuItemBackground(self,*args):
  """
  OnRenderMenuItemBackground(self: ToolStripRenderer,e: ToolStripItemRenderEventArgs)

   Raises the System.Windows.Forms.ToolStripRenderer.RenderMenuItemBackground event.

  

   e: A System.Windows.Forms.ToolStripItemRenderEventArgs that contains the event data.
  """
  pass
 def OnRenderOverflowButtonBackground(self,*args):
  """
  OnRenderOverflowButtonBackground(self: ToolStripRenderer,e: ToolStripItemRenderEventArgs)

   Raises the System.Windows.Forms.ToolStripRenderer.RenderOverflowButtonBackground event.

  

   e: A System.Windows.Forms.ToolStripItemRenderEventArgs that contains the event data.
  """
  pass
 def OnRenderSeparator(self,*args):
  """
  OnRenderSeparator(self: ToolStripRenderer,e: ToolStripSeparatorRenderEventArgs)

   Raises the System.Windows.Forms.ToolStripRenderer.RenderSeparator event.

  

   e: A System.Windows.Forms.ToolStripSeparatorRenderEventArgs that contains the event data.
  """
  pass
 def OnRenderSplitButtonBackground(self,*args):
  """
  OnRenderSplitButtonBackground(self: ToolStripRenderer,e: ToolStripItemRenderEventArgs)

   Raises the 

    System.Windows.Forms.ToolStripRenderer.OnRenderSplitButtonBackground(System.Windows.Forms.ToolStr

    ipItemRenderEventArgs) event.

  

  

   e: A System.Windows.Forms.ToolStripItemRenderEventArgs that contains the event data.
  """
  pass
 def OnRenderStatusStripSizingGrip(self,*args):
  """
  OnRenderStatusStripSizingGrip(self: ToolStripRenderer,e: ToolStripRenderEventArgs)

   Raises the System.Windows.Forms.ToolStripRenderer.RenderStatusStripSizingGrip event.

  

   e: A System.Windows.Forms.ToolStripRenderEventArgs that contains the event data.
  """
  pass
 def OnRenderToolStripBackground(self,*args):
  """
  OnRenderToolStripBackground(self: ToolStripRenderer,e: ToolStripRenderEventArgs)

   Raises the System.Windows.Forms.ToolStripRenderer.RenderToolStripBackground event.

  

   e: A System.Windows.Forms.ToolStripRenderEventArgs that contains the event data.
  """
  pass
 def OnRenderToolStripBorder(self,*args):
  """
  OnRenderToolStripBorder(self: ToolStripRenderer,e: ToolStripRenderEventArgs)

   Raises the System.Windows.Forms.ToolStripRenderer.RenderToolStripBorder event.

  

   e: A System.Windows.Forms.ToolStripRenderEventArgs that contains the event data.
  """
  pass
 def OnRenderToolStripContentPanelBackground(self,*args):
  """
  OnRenderToolStripContentPanelBackground(self: ToolStripRenderer,e: ToolStripContentPanelRenderEventArgs)

   Raises the System.Windows.Forms.ToolStripRenderer.RenderToolStripContentPanelBackground event.

  

   e: A System.Windows.Forms.ToolStripContentPanelRenderEventArgs that contains the event data.
  """
  pass
 def OnRenderToolStripPanelBackground(self,*args):
  """
  OnRenderToolStripPanelBackground(self: ToolStripRenderer,e: ToolStripPanelRenderEventArgs)

   Raises the System.Windows.Forms.ToolStripRenderer.RenderToolStripPanelBackground event.

  

   e: A System.Windows.Forms.ToolStripPanelRenderEventArgs that contains the event data.
  """
  pass
 def OnRenderToolStripStatusLabelBackground(self,*args):
  """
  OnRenderToolStripStatusLabelBackground(self: ToolStripRenderer,e: ToolStripItemRenderEventArgs)

   Raises the System.Windows.Forms.ToolStripRenderer.RenderToolStripStatusLabelBackground event.

  

   e: A System.Windows.Forms.ToolStripItemRenderEventArgs that contains the event data.
  """
  pass
 def ScaleArrowOffsetsIfNeeded(self,*args):
  """ ScaleArrowOffsetsIfNeeded() """
  pass
 Offset2X=2
 Offset2Y=2
 RenderArrow=None
 RenderButtonBackground=None
 RenderDropDownButtonBackground=None
 RenderGrip=None
 RenderImageMargin=None
 RenderItemBackground=None
 RenderItemCheck=None
 RenderItemImage=None
 RenderItemText=None
 RenderLabelBackground=None
 RenderMenuItemBackground=None
 RenderOverflowButtonBackground=None
 RenderSeparator=None
 RenderSplitButtonBackground=None
 RenderStatusStripSizingGrip=None
 RenderToolStripBackground=None
 RenderToolStripBorder=None
 RenderToolStripContentPanelBackground=None
 RenderToolStripPanelBackground=None
 RenderToolStripStatusLabelBackground=None




class DefaultBindingPropertyAttribute(Attribute,_Attribute):
 """
 Specifies the default binding property for a component. This class cannot be inherited.

 

 DefaultBindingPropertyAttribute()

 DefaultBindingPropertyAttribute(name: str)
 """
 def Equals(self,obj):
  """
  Equals(self: DefaultBindingPropertyAttribute,obj: object) -> bool

  

   Determines whether the specified System.Object is equal to the current 

    System.ComponentModel.DefaultBindingPropertyAttribute instance.

  

  

   obj: The System.Object to compare with the current 

    System.ComponentModel.DefaultBindingPropertyAttribute instance

  

   Returns: true if the object is equal to the current instance; otherwise,false,indicating they are not 

    equal.
  """
  pass
 def GetHashCode(self):
  """
  GetHashCode(self: DefaultBindingPropertyAttribute) -> int

   Returns: A 32-bit signed integer hash code.
  """
  pass
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==y """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,name=None):
  """
  __new__(cls: type)

  __new__(cls: type,name: str)
  """
  pass
 def __ne__(self,*args):
  pass
 Name=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the name of the default binding property for the component to which the System.ComponentModel.DefaultBindingPropertyAttribute is bound.



Get: Name(self: DefaultBindingPropertyAttribute) -> str



"""


 Default=None




class ICertificatePolicy:
 """ Validates a server certificate. """
 def CheckValidationResult(self,srvPoint,certificate,request,certificateProblem):
  """
  CheckValidationResult(self: ICertificatePolicy,srvPoint: ServicePoint,certificate: X509Certificate,request: WebRequest,certificateProblem: int) -> bool

  

   Validates a server certificate.

  

   srvPoint: The System.Net.ServicePoint that will use the certificate.

   certificate: The certificate to validate.

   request: The request that received the certificate.

   certificateProblem: The problem that was encountered when using the certificate.

   Returns: true if the certificate should be honored; otherwise,false.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass



class CalendarDateRange(object,INotifyPropertyChanged):
 """
 Represents a range of dates in a System.Windows.Controls.Calendar.

 

 CalendarDateRange()

 CalendarDateRange(day: DateTime)

 CalendarDateRange(start: DateTime,end: DateTime)
 """
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,*__args):
  """
  __new__(cls: type)

  __new__(cls: type,day: DateTime)

  __new__(cls: type,start: DateTime,end: DateTime)
  """
  pass
 def __repr__(self,*args):
  """ __repr__(self: object) -> str """
  pass
 End=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the last date in the represented range.



Get: End(self: CalendarDateRange) -> DateTime



Set: End(self: CalendarDateRange)=value

"""

 Start=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the first date in the represented range.



Get: Start(self: CalendarDateRange) -> DateTime



Set: Start(self: CalendarDateRange)=value

"""


 PropertyChanged=None




class Lazy(object):
 """
 Lazy[T]()

 Lazy[T](valueFactory: Func[T])

 Lazy[T](isThreadSafe: bool)

 Lazy[T](mode: LazyThreadSafetyMode)

 Lazy[T](valueFactory: Func[T],isThreadSafe: bool)

 Lazy[T](valueFactory: Func[T],mode: LazyThreadSafetyMode)
 """
 def ToString(self):
  """
  ToString(self: Lazy[T]) -> str

  

   Creates and returns a string representation of the System.Lazy property for this instance.

   Returns: The result of calling the System.Object.ToString method on the System.Lazy property for this 

    instance,if the value has been created (that is,if the System.Lazy property returns true). 

    Otherwise,a string indicating that the value has not been created.
  """
  pass
 @staticmethod
 def __new__(self,*__args):
  """
  __new__(cls: type)

  __new__(cls: type,valueFactory: Func[T])

  __new__(cls: type,isThreadSafe: bool)

  __new__(cls: type,mode: LazyThreadSafetyMode)

  __new__(cls: type,valueFactory: Func[T],isThreadSafe: bool)

  __new__(cls: type,valueFactory: Func[T],mode: LazyThreadSafetyMode)
  """
  pass
 IsValueCreated=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value that indicates whether a value has been created for this System.Lazy instance.



Get: IsValueCreated(self: Lazy[T]) -> bool



"""

 Value=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the lazily initialized value of the current System.Lazy instance.



Get: Value(self: Lazy[T]) -> T



"""





class TableLayoutPanelGrowStyle(Enum,IComparable,IFormattable,IConvertible):
 """
 Specifies how a System.Windows.Forms.TableLayoutPanel will gain additional rows or columns after its existing cells are full.

 

 enum TableLayoutPanelGrowStyle,values: AddColumns (2),AddRows (1),FixedSize (0)
 """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 AddColumns=None
 AddRows=None
 FixedSize=None
 value__=None




class ToolboxItemFilterAttribute(Attribute,_Attribute):
 """
 Specifies the filter string and filter type to use for a toolbox item.

 

 ToolboxItemFilterAttribute(filterString: str)

 ToolboxItemFilterAttribute(filterString: str,filterType: ToolboxItemFilterType)
 """
 def Equals(self,obj):
  """
  Equals(self: ToolboxItemFilterAttribute,obj: object) -> bool

  

   obj: The object to compare.
  """
  pass
 def GetHashCode(self):
  """ GetHashCode(self: ToolboxItemFilterAttribute) -> int """
  pass
 def Match(self,obj):
  """
  Match(self: ToolboxItemFilterAttribute,obj: object) -> bool

  

   Indicates whether the specified object has a matching filter string.

  

   obj: The object to test for a matching filter string.

   Returns: true if the specified object has a matching filter string; otherwise,false.
  """
  pass
 def ToString(self):
  """ ToString(self: ToolboxItemFilterAttribute) -> str """
  pass
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==y """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,filterString,filterType=None):
  """
  __new__(cls: type,filterString: str)

  __new__(cls: type,filterString: str,filterType: ToolboxItemFilterType)
  """
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 FilterString=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the filter string for the toolbox item.



Get: FilterString(self: ToolboxItemFilterAttribute) -> str



"""

 FilterType=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the type of the filter.



Get: FilterType(self: ToolboxItemFilterAttribute) -> ToolboxItemFilterType



"""

 TypeId=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the type ID for the attribute.



Get: TypeId(self: ToolboxItemFilterAttribute) -> object



"""





class CorrelationManager(object):
 """ Correlates traces that are part of a logical transaction. """
 def StartLogicalOperation(self,operationId=None):
  """
  StartLogicalOperation(self: CorrelationManager)

   Starts a logical operation on a thread.

  StartLogicalOperation(self: CorrelationManager,operationId: object)

   Starts a logical operation with the specified identity on a thread.

  

   operationId: An object identifying the operation.
  """
  pass
 def StopLogicalOperation(self):
  """
  StopLogicalOperation(self: CorrelationManager)

   Stops the current logical operation.
  """
  pass
 ActivityId=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the identity for a global activity.



Get: ActivityId(self: CorrelationManager) -> Guid



Set: ActivityId(self: CorrelationManager)=value

"""

 LogicalOperationStack=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the logical operation stack from the call context.



Get: LogicalOperationStack(self: CorrelationManager) -> Stack



"""





class Debugger(object):
 """
 Enables communication with a debugger. This class cannot be inherited.

 

 Debugger()
 """
 @staticmethod
 def Break():
  """
  Break()

   Signals a breakpoint to an attached debugger.
  """
  pass
 @staticmethod
 def IsLogging():
  """
  IsLogging() -> bool

  

   Checks to see if logging is enabled by an attached debugger.

   Returns: true if a debugger is attached and logging is enabled; otherwise,false. The attached debugger 

    is the registered managed debugger in the DbgManagedDebugger registry key. For more information 

    on this key,see Enabling JIT-Attach Debugging.
  """
  pass
 @staticmethod
 def Launch():
  """
  Launch() -> bool

  

   Launches and attaches a debugger to the process.

   Returns: true if the startup is successful or if the debugger is already attached; otherwise,false.
  """
  pass
 @staticmethod
 def Log(level,category,message):
  """
  Log(level: int,category: str,message: str)

   Posts a message for the attached debugger.

  

   level: A description of the importance of the message.

   category: The category of the message.

   message: The message to show.
  """
  pass
 @staticmethod
 def NotifyOfCrossThreadDependency():
  """
  NotifyOfCrossThreadDependency()

   Notifies a debugger that execution is about to enter a path that involves a cross-thread 

    dependency.
  """
  pass
 DefaultCategory=None
 IsAttached=False




class Timer(Component,IComponent,IDisposable):
 """
 Implements a timer that raises an event at user-defined intervals. This timer is optimized for use in Windows Forms applications and must be used in a window.

 

 Timer()

 Timer(container: IContainer)
 """
 def Dispose(self):
  """
  Dispose(self: Timer,disposing: bool)

   Disposes of the resources,other than memory,used by the timer.

  

   disposing: true to release both managed and unmanaged resources. false to release only the unmanaged 

    resources.
  """
  pass
 def GetService(self,*args):
  """
  GetService(self: Component,service: Type) -> object

  

   Returns an object that represents a service provided by the System.ComponentModel.Component or 

    by its System.ComponentModel.Container.

  

  

   service: A service provided by the System.ComponentModel.Component.

   Returns: An System.Object that represents a service provided by the System.ComponentModel.Component,or 

    null if the System.ComponentModel.Component does not provide the specified service.
  """
  pass
 def MemberwiseClone(self,*args):
  """
  MemberwiseClone(self: MarshalByRefObject,cloneIdentity: bool) -> MarshalByRefObject

  

   Creates a shallow copy of the current System.MarshalByRefObject object.

  

   cloneIdentity: false to delete the current System.MarshalByRefObject object's identity,which will cause the 

    object to be assigned a new identity when it is marshaled across a remoting boundary. A value of 

    false is usually appropriate. true to copy the current System.MarshalByRefObject object's 

    identity to its clone,which will cause remoting client calls to be routed to the remote server 

    object.

  

   Returns: A shallow copy of the current System.MarshalByRefObject object.

  MemberwiseClone(self: object) -> object

  

   Creates a shallow copy of the current System.Object.

   Returns: A shallow copy of the current System.Object.
  """
  pass
 def OnTick(self,*args):
  """
  OnTick(self: Timer,e: EventArgs)

   Raises the System.Windows.Forms.Timer.Tick event.

  

   e: An System.EventArgs that contains the event data. This is always System.EventArgs.Empty.
  """
  pass
 def Start(self):
  """
  Start(self: Timer)

   Starts the timer.
  """
  pass
 def Stop(self):
  """
  Stop(self: Timer)

   Stops the timer.
  """
  pass
 def ToString(self):
  """
  ToString(self: Timer) -> str

  

   Returns a string that represents the System.Windows.Forms.Timer.

   Returns: A string that represents the current System.Windows.Forms.Timer.
  """
  pass
 def __enter__(self,*args):
  """
  __enter__(self: IDisposable) -> object

  

   Provides the implementation of __enter__ for objects which implement IDisposable.
  """
  pass
 def __exit__(self,*args):
  """
  __exit__(self: IDisposable,exc_type: object,exc_value: object,exc_back: object)

   Provides the implementation of __exit__ for objects which implement IDisposable.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,container=None):
  """
  __new__(cls: type)

  __new__(cls: type,container: IContainer)
  """
  pass
 def __str__(self,*args):
  pass
 CanRaiseEvents=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the component can raise an event.



"""

 DesignMode=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value that indicates whether the System.ComponentModel.Component is currently in design mode.



"""

 Enabled=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets whether the timer is running.



Get: Enabled(self: Timer) -> bool



Set: Enabled(self: Timer)=value

"""

 Events=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the list of event handlers that are attached to this System.ComponentModel.Component.



"""

 Interval=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the time,in milliseconds,before the System.Windows.Forms.Timer.Tick event is raised relative to the last occurrence of the System.Windows.Forms.Timer.Tick event.



Get: Interval(self: Timer) -> int



Set: Interval(self: Timer)=value

"""

 Tag=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets an arbitrary string representing some type of user state.



Get: Tag(self: Timer) -> object



Set: Tag(self: Timer)=value

"""


 Tick=None




class VarEnum(Enum,IComparable,IFormattable,IConvertible):
 """
 Indicates how to marshal the array elements when an array is marshaled from managed to unmanaged code as a System.Runtime.InteropServices.UnmanagedType.SafeArray.

 

 enum VarEnum,values: VT_ARRAY (8192),VT_BLOB (65),VT_BLOB_OBJECT (70),VT_BOOL (11),VT_BSTR (8),VT_BYREF (16384),VT_CARRAY (28),VT_CF (71),VT_CLSID (72),VT_CY (6),VT_DATE (7),VT_DECIMAL (14),VT_DISPATCH (9),VT_EMPTY (0),VT_ERROR (10),VT_FILETIME (64),VT_HRESULT (25),VT_I1 (16),VT_I2 (2),VT_I4 (3),VT_I8 (20),VT_INT (22),VT_LPSTR (30),VT_LPWSTR (31),VT_NULL (1),VT_PTR (26),VT_R4 (4),VT_R8 (5),VT_RECORD (36),VT_SAFEARRAY (27),VT_STORAGE (67),VT_STORED_OBJECT (69),VT_STREAM (66),VT_STREAMED_OBJECT (68),VT_UI1 (17),VT_UI2 (18),VT_UI4 (19),VT_UI8 (21),VT_UINT (23),VT_UNKNOWN (13),VT_USERDEFINED (29),VT_VARIANT (12),VT_VECTOR (4096),VT_VOID (24)
 """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 value__=None
 VT_ARRAY=None
 VT_BLOB=None
 VT_BLOB_OBJECT=None
 VT_BOOL=None
 VT_BSTR=None
 VT_BYREF=None
 VT_CARRAY=None
 VT_CF=None
 VT_CLSID=None
 VT_CY=None
 VT_DATE=None
 VT_DECIMAL=None
 VT_DISPATCH=None
 VT_EMPTY=None
 VT_ERROR=None
 VT_FILETIME=None
 VT_HRESULT=None
 VT_I1=None
 VT_I2=None
 VT_I4=None
 VT_I8=None
 VT_INT=None
 VT_LPSTR=None
 VT_LPWSTR=None
 VT_NULL=None
 VT_PTR=None
 VT_R4=None
 VT_R8=None
 VT_RECORD=None
 VT_SAFEARRAY=None
 VT_STORAGE=None
 VT_STORED_OBJECT=None
 VT_STREAM=None
 VT_STREAMED_OBJECT=None
 VT_UI1=None
 VT_UI2=None
 VT_UI4=None
 VT_UI8=None
 VT_UINT=None
 VT_UNKNOWN=None
 VT_USERDEFINED=None
 VT_VARIANT=None
 VT_VECTOR=None
 VT_VOID=None




class QueryContinueDragEventArgs(EventArgs):
 """
 Provides data for the System.Windows.Forms.Control.QueryContinueDrag event.

 

 QueryContinueDragEventArgs(keyState: int,escapePressed: bool,action: DragAction)
 """
 @staticmethod
 def __new__(self,keyState,escapePressed,action):
  """ __new__(cls: type,keyState: int,escapePressed: bool,action: DragAction) """
  pass
 Action=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the status of a drag-and-drop operation.



Get: Action(self: QueryContinueDragEventArgs) -> DragAction



Set: Action(self: QueryContinueDragEventArgs)=value

"""

 EscapePressed=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets whether the user pressed the ESC key.



Get: EscapePressed(self: QueryContinueDragEventArgs) -> bool



"""

 KeyState=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the current state of the SHIFT,CTRL,and ALT keys.



Get: KeyState(self: QueryContinueDragEventArgs) -> int



"""





# -*- coding: utf-8 -*-
"""Routes HTTP Server."""
#pylint: disable=import-error,invalid-name,broad-except
#pylint: disable=missing-docstring
import sys
import traceback
import cgi
import json
import threading

from pyrevit.api import UI
from pyrevit.coreutils.logger import get_logger
from pyrevit.compat import PY3
from pyrevit.compat import urlparse

from pyrevit.routes.server import exceptions as excp
from pyrevit.routes.server import base
from pyrevit.routes.server import handler
from pyrevit.routes.server import router

if PY3:
    from http.server import BaseHTTPRequestHandler, HTTPServer
    from socketserver import ThreadingMixIn
else:
    from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
    from SocketServer import ThreadingMixIn


mlogger = get_logger(__name__)


# instance of event handler created when this module is loaded
# on hosts main thread. Creating external events on non-main threads
# are prohibited by the host. this event handler is reconfigured
# for every request registered by this module
REQUEST_HNDLR = handler.RequestHandler()
EVENT_HNDLR = UI.ExternalEvent.Create(REQUEST_HNDLR)


class HttpRequestHandler(BaseHTTPRequestHandler):
    """HTTP Requests Handler."""
    def _parse_api_path(self):
        url_parts = urlparse(self.path)
        if url_parts:
            levels = url_parts.path.split('/')
            # host:ip/<api_name>/<route>/.../.../...
            if levels and len(levels) >= 2:
                api_name = levels[1]
                if len(levels) > 2:
                    api_path = '/' + '/'.join(levels[2:])
                else:
                    api_path = '/'
                return api_name, api_path
        return None, None

    def _parse_request_info(self):
        # find the app
        api_name, api_path = self._parse_api_path() #type:str, str
        if not api_name:
            raise excp.APINotDefinedException(api_name)
        return api_name, api_path

    def _find_route_handler(self, api_name, path, method):
        route, route_handler = router.get_route_handler(
            api_name=api_name,
            path=path,
            method=method
            )
        if not route_handler:
            raise excp.RouteHandlerNotDefinedException(api_name, path, method)
        return route, route_handler

    def _prepare_request(self, route, path, method):
        # process request data
        data = None
        content_length = self.headers.getheader('content-length') # type: str
        if content_length and content_length.isnumeric():
            data = self.rfile.read(int(content_length))
            # format data
            content_type_header = self.headers.getheader('content-type')
            if content_type_header:
                content_type, _ = cgi.parse_header(content_type_header)
                if content_type == 'application/json':
                    data = json.loads(data)

        return base.Request(
            path=path,
            method=method,
            data=data,
            params=router.extract_route_params(route.pattern, path)
        )

    def _prepare_host_handler(self, request, route_handler):
        # create the base Revit external event handler
        # upon Raise(), finds and runs the appropriate func
        REQUEST_HNDLR.request = request
        REQUEST_HNDLR.handler = route_handler
        return REQUEST_HNDLR, EVENT_HNDLR

    def _call_host_event(self, req_hndlr, event_hndlr):
        # reset handler
        req_hndlr.reset()
        # raise request to host
        extevent_raise_response = event_hndlr.Raise()
        if extevent_raise_response == UI.ExternalEventRequest.Denied:
            raise excp.RouteHandlerDeniedException(req_hndlr.request)
        elif extevent_raise_response == UI.ExternalEventRequest.TimedOut:
            raise excp.RouteHandlerTimedOutException(req_hndlr.request)

    def _call_host_event_sync(self, req_hndlr, event_hndlr):
        # call handler
        self._call_host_event(req_hndlr, event_hndlr)

        # wait until event has been picked up by host for execution
        while event_hndlr.IsPending:
            pass

        # wait until handler signals completion
        req_hndlr.join()

    def _write_response(self, response):
        r = handler.RequestHandler.parse_response(response)
        self.send_response(r.status)
        if r.headers:
            for key, value in r.headers.items():
                self.send_header(key, value)
            self.end_headers()
        # sending \n if no data otherwise Postman panics for some reason
        self.wfile.write(r.data or "\n")

    def _handle_route(self, method):
        # process the given url and find API and route
        api_name, path = self._parse_request_info()

        # find the handler function registered by the API and route
        route, route_handler = self._find_route_handler(api_name, path, method)

        # prepare a request obj to be passed to registered handler
        request = self._prepare_request(route, path, method)

        # if handler has uiapp in arguments, run in host api context
        if handler.RequestHandler.wants_api_context(route_handler):
            # create a handler and event object in host
            req_hndlr, event_hndlr = \
                self._prepare_host_handler(request, route_handler)

            # do the handling work
            # if request has callback url, raise the event handler and return
            #   the handler, when executed, will notify the callback url
            if request.callback_url:
                self._call_host_event(req_hndlr, event_hndlr)
                # acknowledge the request is accepted and return
                self._write_response(base.Response(status=base.NO_CONTENT))
            # otherwise run the handler and wait
            else:
                self._call_host_event_sync(req_hndlr, event_hndlr)
                # prepare response
                # grab response from req_hndlr.response
                # req_hndlr.response getter is thread-safe
                self._write_response(req_hndlr.response)
        # otherwise run here
        else:
            # now run the method, and gret response
            response = \
                handler.RequestHandler.run_handler(
                    handler=route_handler,
                    kwargs=handler.RequestHandler.prepare_handler_kwargs(
                        request=request,
                        handler=route_handler
                    )
                )
            # prepare response
            self._write_response(response)

    def _process_request(self, method):
        # this method is wrapping the actual handler and is
        # catching all the excp
        try:
            self._handle_route(method=method)
        except Exception as ex:
            # get exception info
            sys.exc_type, sys.exc_value, sys.exc_traceback = \
                sys.exc_info()
            # go back one frame to grab exception stack from handler
            # and grab traceback lines
            tb_report = ''.join(
                traceback.format_tb(sys.exc_traceback)[1:]
            )
            self._write_response(
                excp.ServerException(
                    message=str(ex),
                    exception_type=sys.exc_type,
                    exception_traceback=tb_report
                )
            )

    # CRUD Methods ------------------------------------------------------------
    # create
    def do_POST(self):
        self._process_request(method='POST')

    # read
    def do_GET(self):
        self._process_request(method='GET')

    # update
    def do_PUT(self):
        self._process_request(method='PUT')

    # delete
    def do_DELETE(self):
        self._process_request(method='DELETE')

    # rest of standard http methods -------------------------------------------
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods
    def do_HEAD(self):
        self._process_request(method='HEAD')

    def do_CONNECT(self):
        self._process_request(method='CONNECT')

    def do_OPTIONS(self):
        self._process_request(method='OPTIONS')

    def do_TRACE(self):
        self._process_request(method='TRACE')

    def do_PATCH(self):
        self._process_request(method='PATCH')


class ThreadedHttpServer(ThreadingMixIn, HTTPServer):
    """Threaded HTTP server."""
    allow_reuse_address = True

    def shutdown(self):
        self.socket.close()
        HTTPServer.shutdown(self)


class RoutesServer(object):
    """Route server thread handler.

    It runs an HTTP server on the given host and port.

    Args:
        host (str): host
        port (int): port
    """
    def __init__(self, host, port):
        self.server = ThreadedHttpServer((host, port), HttpRequestHandler)
        self.host = host
        self.port = port
        self.start()

    def __str__(self):
        return "Routes server is listening on http://%s:%s" \
            % (self.host or "0.0.0.0", self.port)

    def __repr__(self):
        return '<RoutesServer @ http://%s:%s>' \
            % (self.host or "0.0.0.0", self.port)

    def start(self):
        self.server_thread = threading.Thread(target=self.server.serve_forever)
        self.server_thread.daemon = True
        self.server_thread.start()

    def waitForThread(self):
        self.server_thread.join()

    def stop(self):
        self.server.shutdown()
        self.waitForThread()



class BorderGapMaskConverter(object,IMultiValueConverter):
 """
 Represents a converter that converts the dimensions of a System.Windows.Controls.GroupBox control into a System.Windows.Media.VisualBrush.

 

 BorderGapMaskConverter()
 """
 def Convert(self,values,targetType,parameter,culture):
  """
  Convert(self: BorderGapMaskConverter,values: Array[object],targetType: Type,parameter: object,culture: CultureInfo) -> object

  

   Creates a System.Windows.Media.VisualBrush that draws the border for a 

    System.Windows.Controls.GroupBox control.

  

  

   values: An array of three numbers that represent the System.Windows.Controls.GroupBox control 

    parameters.  See Remarks for more information.

  

   targetType: This parameter is not used.

   parameter: The width of the visible line to the left of the 

    System.Windows.Controls.HeaderedContentControl.Header in the System.Windows.Controls.GroupBox.

  

   culture: This parameter is not used.

   Returns: A System.Windows.Media.VisualBrush that draws the border around a 

    System.Windows.Controls.GroupBox control that includes a gap for the 

    System.Windows.Controls.HeaderedContentControl.Header content.
  """
  pass
 def ConvertBack(self,value,targetTypes,parameter,culture):
  """
  ConvertBack(self: BorderGapMaskConverter,value: object,targetTypes: Array[Type],parameter: object,culture: CultureInfo) -> Array[object]

  

   Not implemented.

  

   value: This parameter is not used.

   targetTypes: This parameter is not used.

   parameter: This parameter is not used.

   culture: This parameter is not used.

   Returns: System.Windows.Data.Binding.DoNothing in all cases.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __repr__(self,*args):
  """ __repr__(self: object) -> str """
  pass



"""Base module for pushing toast messages on Win 10.

This module is a wrapper for a cli utility that provides toast message
functionality. See `https://github.com/go-toast/toast`
"""

import os.path as op
import subprocess

from pyrevit import ROOT_BIN_DIR
from pyrevit.coreutils.logger import get_logger


#pylint: disable=W0703,C0302
mlogger = get_logger(__name__)  #pylint: disable=C0103


def get_toaster():
    """Return full file path of the toast binary utility."""
    return op.join(op.dirname(__file__), 'pyrevit-toast.exe')


def send_toast(message,
               title=None, appid=None, icon=None, click=None, actions=None):
    """Send toast notificaton.

    Args:
        message (str): notification message
        title (str): notification title
        appid (str): application unique id (see `--app-id` cli option)
        icon (str): notification icon (see `--icon` cli option)
        click (str): click action (see `--activation-arg` cli option)
        actions (dict[str:str]):
            list of actions (see `--action` and `--action-arg` cli options)
    """
    # set defaults
    if not title:
        title = 'pyRevit'
    if not appid:
        appid = title
    if not icon:
        icon = op.join(ROOT_BIN_DIR, 'pyRevit.ico')
    if not actions:
        actions = {}

    # build the toast
    toast_args = r'"{}"'.format(get_toaster())
    toast_args += r' --app-id "{}"'.format(appid)
    toast_args += r' --title "{}"'.format(title)
    toast_args += r' --message "{}"'.format(message)
    toast_args += r' --icon "{}"'.format(icon)
    toast_args += r' --audio "default"'
    # toast_args += r' --duration "long"'
    if click:
        toast_args += r' --activation-arg "{}"'.format(click)
    for action, args in actions.items():
        toast_args += r' --action "{}" --action-arg "{}"'.format(action, args)

    # send the toast now
    mlogger.debug('toasting: %s', toast_args)
    subprocess.Popen(toast_args, shell=True)



class ClassInterfaceAttribute(Attribute,_Attribute):
 """
 Indicates the type of class interface to be generated for a class exposed to COM,if an interface is generated at all.

 

 ClassInterfaceAttribute(classInterfaceType: ClassInterfaceType)

 ClassInterfaceAttribute(classInterfaceType: Int16)
 """
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,classInterfaceType):
  """
  __new__(cls: type,classInterfaceType: ClassInterfaceType)

  __new__(cls: type,classInterfaceType: Int16)
  """
  pass
 Value=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the System.Runtime.InteropServices.ClassInterfaceType value that describes which type of interface should be generated for the class.



Get: Value(self: ClassInterfaceAttribute) -> ClassInterfaceType



"""





class QuestionEventArgs(EventArgs):
 """
 Provides data for events that need a true or false answer to a question.

 

 QuestionEventArgs()

 QuestionEventArgs(response: bool)
 """
 @staticmethod
 def __new__(self,response=None):
  """
  __new__(cls: type)

  __new__(cls: type,response: bool)
  """
  pass
 Response=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a value indicating the response to a question represented by the event.



Get: Response(self: QuestionEventArgs) -> bool



Set: Response(self: QuestionEventArgs)=value

"""





"""Unit tests facility."""
import time
from unittest import TestResult, TestLoader

from pyrevit.coreutils.logger import get_logger
from pyrevit.output import get_output


#pylint: disable=W0703,C0302,C0103
mlogger = get_logger(__name__)


DEBUG_OKAY_RESULT = 'PASSED'
DEBUG_FAIL_RESULT = 'FAILED'

RESULT_TEST_SUITE_START = '<div class="unittest unitteststart">' \
                          'Test Suite: {suite}' \
                          '</div>'

RESULT_DIV_OKAY = '<div class="unittest unittestokay">' \
                  ':white_heavy_check_mark: PASSED {test}' \
                  '</div>'

RESULT_DIV_FAIL = '<div class="unittest unittestfail">' \
                  ':cross_mark: FAILED {test}' \
                  '</div>'

RESULT_DIV_ERROR = '<div class="unittest unittesterror">' \
                   ':heavy_large_circle: ERROR {test}' \
                   '</div>'


class OutputWriter:
    """Output writer for tests results."""
    def __init__(self):
        self._output = get_output()

    def write(self, output_str):
        """Prints the results to the output window.

        Args:
            output_str (str): Text to output
        """
        self._output.print_html(output_str)


class PyRevitTestResult(TestResult):
    """Pyrevit Test Result.

    Args:
        verbosity (int): verbosity level.
    """
    def __init__(self, verbosity):
        super(PyRevitTestResult, self).__init__(verbosity=verbosity)
        self.writer = OutputWriter()

    @staticmethod
    def getDescription(test):
        """Returns the description of the test.

        Args:
            test (TestCase): Unit test.

        Returns:
            (str): test description
        """
        return test.shortDescription() or test

    def startTest(self, test):
        """Starts the test.

        Args:
            test (TestCase): unit test
        """
        super(PyRevitTestResult, self).startTest(test)
        mlogger.debug('Running test: %s', self.getDescription(test))

    def addSuccess(self, test):
        """Adds a test success.

        Args:
            test (TestCase): unit test case
        """
        super(PyRevitTestResult, self).addSuccess(test)
        mlogger.debug(DEBUG_OKAY_RESULT)
        self.writer.write(RESULT_DIV_OKAY
                          .format(test=self.getDescription(test)))

    def addError(self, test, err):
        """Adds a test error.

        Args:
            test (TestCase): unit test case
            err (OptExcInfo): test exception info
        """
        super(PyRevitTestResult, self).addError(test, err)
        mlogger.debug(DEBUG_FAIL_RESULT)
        self.writer.write(RESULT_DIV_ERROR
                          .format(test=self.getDescription(test)))

    def addFailure(self, test, err):
        """Adds a test failure.

        Args:
            test (TestCase): unit test case
            err (OptExcInfo): test exception info
        """
        super(PyRevitTestResult, self).addFailure(test, err)
        mlogger.debug(DEBUG_FAIL_RESULT)
        self.writer.write(RESULT_DIV_FAIL
                          .format(test=self.getDescription(test)))

    # def addSkip(self, test, reason):
    #     super(PyRevitTestResult, self).addSkip(test, reason)

    # def addExpectedFailure(self, test, err):
    #     super(PyRevitTestResult, self).addExpectedFailure(test, err)

    # def addUnexpectedSuccess(self, test):
    #     super(PyRevitTestResult, self).addUnexpectedSuccess(test)


class PyRevitTestRunner(object):
    """Test runner.

    Args:
        verbosity (int): level of vermosity. Defaults to 1.
        failfast (bool): if True, stops at the first failure. Defaults to False.
        use_buffer (bool): use a buffer. Defaults to False.
        resultclass (type): Class to use to hold the results. 
            Defaults to `PyRevitTestResult`.
    """
    resultclass = PyRevitTestResult

    def __init__(self, verbosity=1, failfast=False,
                 use_buffer=False, resultclass=None):
        self.verbosity = verbosity
        self.failfast = failfast
        self.use_buffer = use_buffer
        if resultclass is not None:
            self.resultclass = resultclass

    def _make_result(self):
        return self.resultclass(self.verbosity)

    def run(self, test):
        """Runs a test suite.

        Args:
            test (TestSuite): Test suite to run

        Returns:
            (PyRevitTestResult): Test suite results.
        """
        # setup results object
        result = self._make_result()
        result.failfast = self.failfast
        result.buffer = self.use_buffer

        # start clock
        start_time = time.time()

        # find run test methods
        start_test_run = getattr(result, 'startTestRun', None)
        if start_test_run is not None:
            start_test_run()
        try:
            test(result)
        finally:
            stop_test_run = getattr(result, 'stopTestRun', None)
            if stop_test_run is not None:
                stop_test_run()

        # stop clock and calculate run time
        stop_time = time.time()
        time_taken = stop_time - start_time

        # print errots
        result.printErrors()
        test_count = result.testsRun
        mlogger.debug("Ran %d test%s in %.3fs",
                      test_count, test_count != 1 and "s" or "", time_taken)

        expected_fails = unexpected_successes = skipped = 0
        try:
            results = map(len, (result.expectedFailures,
                                result.unexpectedSuccesses,
                                result.skipped))
        except AttributeError:
            pass
        else:
            expected_fails, unexpected_successes, skipped = results

        infos = []
        if not result.wasSuccessful():
            mlogger.debug("FAILED")
            failed, errored = map(len, (result.failures, result.errors))
            if failed:
                infos.append("failures=%d" % failed)
            if errored:
                infos.append("errors=%d" % errored)
        else:
            mlogger.debug(DEBUG_OKAY_RESULT)

        if skipped:
            infos.append("skipped=%d" % skipped)
        if expected_fails:
            infos.append("expected failures=%d" % expected_fails)
        if unexpected_successes:
            infos.append("unexpected successes=%d" % unexpected_successes)
        if infos:
            mlogger.debug(" (%s)", (", ".join(infos),))

        return result


def run_module_tests(test_module):
    """Runs the unit tests of the given module.

    Args:
        test_module (module): module with tests

    Returns:
        (PyRevitTestResult): tests results.
    """
    test_runner = PyRevitTestRunner()
    test_loader = TestLoader()
    # load all testcases from the given module into a testsuite
    test_suite = test_loader.loadTestsFromModule(test_module)
    # run the test suite
    mlogger.debug('Running test suite for module: %s', test_module)
    OutputWriter()\
        .write(RESULT_TEST_SUITE_START.format(suite=test_module.__name__))
    return test_runner.run(test_suite)


def run_test_case(test_case):
    """Runs the unit test of the given TestCase class.

    Args:
        test_case (type[TestCase]): TestCase class with tests

    Returns:
        (PyRevitTestResult): tests results.
    """
    test_runner = PyRevitTestRunner()
    suite = TestLoader().loadTestsFromTestCase(test_case)
    OutputWriter()\
        .write(RESULT_TEST_SUITE_START.format(suite=suite.__class__.__name__))
    return test_runner.run(suite)
